#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from typing import Union
import itertools
import numpy
import sympy

from openfermion.ops.operators import (QuadOperator, BosonOperator,
                                       FermionOperator, MajoranaOperator)
from openfermion.ops.representations import (PolynomialTensor,
                                             DiagonalCoulombHamiltonian)

from openfermion.utils.operator_utils import count_qubits


def get_quad_operator(operator, hbar=1.):
    """Convert to QuadOperator.

    Args:
        operator: BosonOperator.
        hbar (float): the value of hbar used in the definition
            of the commutator [q_i, p_j] = i hbar delta_ij.
            By default hbar=1.

    Returns:
        quad_operator: An instance of the QuadOperator class.
    """
    quad_operator = QuadOperator()

    if isinstance(operator, BosonOperator):
        for term, coefficient in operator.terms.items():
            tmp = QuadOperator('', coefficient)
            for i, d in term:
                tmp *= (1./numpy.sqrt(2.*hbar)) \
                    * (QuadOperator(((i, 'q')))
                        + QuadOperator(((i, 'p')), 1j*(-1)**d))
            quad_operator += tmp

    else:
        raise TypeError("Only BosonOperator is currently "
                        "supported for get_quad_operator.")

    return quad_operator


def check_no_sympy(operator):
    """Checks whether a SymbolicOperator contains any
    sympy expressions, which will prevent it being converted
    to a PolynomialTensor or DiagonalCoulombHamiltonian

    Args:
        operator(SymbolicOperator): the operator to be tested
    """
    for key in operator.terms:
        if isinstance(operator.terms[key], sympy.Expr):
            raise TypeError('This conversion is currently not supported ' +
                            'for operators with sympy expressions ' +
                            'as coefficients')


def get_boson_operator(operator, hbar=1.):
    """Convert to BosonOperator.

    Args:
        operator: QuadOperator.
        hbar (float): the value of hbar used in the definition
            of the commutator [q_i, p_j] = i hbar delta_ij.
            By default hbar=1.

    Returns:
        boson_operator: An instance of the BosonOperator class.
    """
    boson_operator = BosonOperator()

    if isinstance(operator, QuadOperator):
        for term, coefficient in operator.terms.items():
            tmp = BosonOperator('', coefficient)
            for i, d in term:
                if d == 'q':
                    coeff = numpy.sqrt(hbar / 2)
                    sign = 1
                elif d == 'p':
                    coeff = -1j * numpy.sqrt(hbar / 2)
                    sign = -1

                tmp *= coeff * (BosonOperator(((i, 0))) + BosonOperator(
                    ((i, 1)), sign))
            boson_operator += tmp

    else:
        raise TypeError("Only QuadOperator is currently "
                        "supported for get_boson_operator.")

    return boson_operator


def get_fermion_operator(operator):
    """Convert to FermionOperator.

    Returns:
        fermion_operator: An instance of the FermionOperator class.
    """
    if isinstance(operator, PolynomialTensor):
        return _polynomial_tensor_to_fermion_operator(operator)
    elif isinstance(operator, DiagonalCoulombHamiltonian):
        return _diagonal_coulomb_hamiltonian_to_fermion_operator(operator)
    elif isinstance(operator, MajoranaOperator):
        return _majorana_operator_to_fermion_operator(operator)
    else:
        raise TypeError('{} cannot be converted to FermionOperator'.format(
            type(operator)))


def _polynomial_tensor_to_fermion_operator(operator):
    fermion_operator = FermionOperator()
    for term in operator:
        fermion_operator += FermionOperator(term, operator[term])
    return fermion_operator


def _diagonal_coulomb_hamiltonian_to_fermion_operator(operator):
    fermion_operator = FermionOperator()
    n_qubits = count_qubits(operator)
    fermion_operator += FermionOperator((), operator.constant)
    for p, q in itertools.product(range(n_qubits), repeat=2):
        fermion_operator += FermionOperator(((p, 1), (q, 0)),
                                            operator.one_body[p, q])
        fermion_operator += FermionOperator(((p, 1), (p, 0), (q, 1), (q, 0)),
                                            operator.two_body[p, q])
    return fermion_operator


def _majorana_operator_to_fermion_operator(majorana_operator):
    fermion_operator = FermionOperator()
    for term, coeff in majorana_operator.terms.items():
        converted_term = _majorana_term_to_fermion_operator(term)
        converted_term *= coeff
        fermion_operator += converted_term
    return fermion_operator


def _majorana_term_to_fermion_operator(term):
    converted_term = FermionOperator(())
    for index in term:
        j, b = divmod(index, 2)
        if b:
            converted_op = FermionOperator((j, 0), -1j)
            converted_op += FermionOperator((j, 1), 1j)
        else:
            converted_op = FermionOperator((j, 0))
            converted_op += FermionOperator((j, 1))
        converted_term *= converted_op
    return converted_term


def get_majorana_operator(
        operator: Union[PolynomialTensor, DiagonalCoulombHamiltonian,
                        FermionOperator]) -> MajoranaOperator:
    """
    Convert to MajoranaOperator.

    Uses the convention of even + odd indexing of Majorana modes derived from
    a fermionic mode:
        fermion annhil.  c_k  -> ( gamma_{2k} + 1.j * gamma_{2k+1} ) / 2
        fermion creation c^_k -> ( gamma_{2k} - 1.j * gamma_{2k+1} ) / 2

    Args:
        operator (PolynomialTensor,
            DiagonalCoulombHamiltonian or
            FermionOperator): Operator to write as Majorana Operator.

    Returns:
        majorana_operator: An instance of the MajoranaOperator class.

    Raises:
        TypeError: If operator is not of PolynomialTensor,
            DiagonalCoulombHamiltonian or FermionOperator.
    """
    if isinstance(operator, FermionOperator):
        return _fermion_operator_to_majorana_operator(operator)
    elif isinstance(operator, (PolynomialTensor, DiagonalCoulombHamiltonian)):
        return _fermion_operator_to_majorana_operator(
            get_fermion_operator(operator))
    raise TypeError('{} cannot be converted to MajoranaOperator'.format(
        type(operator)))


def _fermion_operator_to_majorana_operator(fermion_operator: FermionOperator
                                          ) -> MajoranaOperator:
    """
    Convert FermionOperator to MajoranaOperator.

    Auxiliar function of get_majorana_operator.

    Args:
        fermion_operator (FermionOperator): To convert to MajoranaOperator.

    Returns:
        majorana_operator object.

    Raises:
        TypeError: if input is not a FermionOperator.
    """
    if not isinstance(fermion_operator, FermionOperator):
        raise TypeError('Input a FermionOperator.')

    majorana_operator = MajoranaOperator()
    for term, coeff in fermion_operator.terms.items():
        converted_term = _fermion_term_to_majorana_operator(term)
        converted_term *= coeff
        majorana_operator += converted_term

    return majorana_operator


def _fermion_term_to_majorana_operator(term: tuple) -> MajoranaOperator:
    """
    Convert single terms of FermionOperator to Majorana.
    (Auxiliary function of get_majorana_operator.)

    Convention: even + odd indexing of Majorana modes derived from a
    fermionic mode:
        fermion annhil.  c_k  -> ( gamma_{2k} + 1.j * gamma_{2k+1} ) / 2
        fermion creation c^_k -> ( gamma_{2k} - 1.j * gamma_{2k+1} ) / 2

    Args:
        term (tuple): single FermionOperator term.

    Returns:
        converted_term: single MajoranaOperator term.

    Raises:
        TypeError: if term is a tuple.
    """
    if not isinstance(term, tuple):
        raise TypeError('Term does not have the correct Type.')

    converted_term = MajoranaOperator(())
    for index, action in term:
        converted_op = MajoranaOperator((2 * index,), 0.5)

        if action:
            converted_op += MajoranaOperator((2 * index + 1,), -0.5j)

        else:
            converted_op += MajoranaOperator((2 * index + 1,), 0.5j)

        converted_term *= converted_op

    return converted_term



from itertools import product
import numpy as np
import pytest
from openfermion.third_party.representability._bijections import Bijection, \
    index_index_basis, index_tuple_basis
from openfermion.third_party.representability._namedtensor import Tensor


def test_namedtensor_initialization():
    td = Tensor()
    assert isinstance(td, Tensor)

    data = np.arange(16).reshape((4, 4))
    td = Tensor(name='opdm', tensor=data, basis=index_index_basis(4))
    assert td.name == 'opdm'
    for i in range(4):
        assert td.basis.fwd(i) == i
    assert np.allclose(td.data, data)

    with pytest.raises(TypeError):
        _ = Tensor(name='opdm',
                   tensor=data,
                   basis=dict(zip(range(4), range(4))))

    td = Tensor(name='opdm', tensor=data)
    assert isinstance(td.basis, Bijection)


def test_namedtensor_getdata():
    data = np.arange(16).reshape((4, 4))
    td = Tensor(name='opdm', tensor=data, basis=index_index_basis(4))
    for i, j in product(range(4), repeat=2):
        assert np.isclose(td[i, j], data[i, j])
    blank_td = Tensor(name='opdm')
    with pytest.raises(TypeError):
        _ = blank_td[0, 0]


def test_namedtensor_call():
    """
    make a matrix that has a different basis than indexing
    """
    n = 4
    dim = int(n * (n - 1) / 2)
    geminals = []
    bas = {}
    cnt = 0
    for i in range(4):
        for j in range(i + 1, 4):
            bas[cnt] = (i, j)
            geminals.append((i, j))
            cnt += 1
    rev_bas = dict(zip(bas.values(), bas.keys()))

    rand_mat = np.random.random((dim, dim))
    basis_bijection = index_tuple_basis(geminals)
    test_tensor = Tensor(tensor=rand_mat, basis=basis_bijection)
    assert test_tensor.basis.fwd(0) == (0, 1)
    assert test_tensor.basis.fwd(2) == (0, 3)
    assert test_tensor.basis.rev(test_tensor.basis.fwd(5)) == 5
    assert test_tensor.ndim == 2
    assert test_tensor.dim == dim
    # index into data directly
    assert test_tensor[2, 3] == rand_mat[2, 3]
    # index into data via basis indexing
    assert test_tensor(0, 1, 0, 1) == rand_mat[0, 0]
    assert test_tensor(0, 1, 0, 1) == rand_mat[0, 0]
    assert test_tensor(1, 2, 0, 1) == rand_mat[rev_bas[(1, 2)], rev_bas[(0, 1)]]
    assert test_tensor.index_vectorized(1, 2, 0, 1) == rev_bas[(1, 2)] * dim + \
           rev_bas[(0, 1)]

    # testing iteration over the upper triangle
    for iter_vals in test_tensor.utri_iterator():
        val, [i, j] = iter_vals
        assert val == rand_mat[test_tensor.basis.rev(i),
                               test_tensor.basis.rev(j)]


def test_tensor_index():
    """
    Testing the index_vectorized which is mapped through the basis
    """
    a = np.arange(16).reshape((4, 4), order='C')
    basis = [(0, 0), (0, 1), (1, 0), (1, 1)]
    basis = index_tuple_basis(basis)
    tt = Tensor(tensor=a, basis=basis)
    assert np.allclose(tt.data, a)
    assert tt.size == 16
    assert isinstance(tt.basis, Bijection)
    assert np.isclose(tt.index_vectorized(0, 0, 0, 0), 0)
    assert np.isclose(tt.index_vectorized(0, 0, 0, 1), 1)
    assert np.isclose(tt.index_vectorized(0, 0, 1, 0), 2)
    assert np.isclose(tt.index_vectorized(0, 0, 1, 1), 3)
    assert np.isclose(tt.index_vectorized(0, 1, 0, 0), 4)
    assert np.isclose(tt.index_vectorized(0, 1, 0, 1), 5)
    assert np.isclose(tt.index_vectorized(0, 1, 1, 0), 6)
    assert np.isclose(tt.index_vectorized(0, 1, 1, 1), 7)
    assert np.isclose(tt.index_vectorized(1, 0, 0, 0), 8)
    # etc...

    a = np.arange(16).reshape((4, 4), order='C')
    tt = Tensor(tensor=a)  # the canonical basis
    assert np.isclose(tt.index_vectorized(0, 0), 0)
    assert np.isclose(tt.index_vectorized(0, 1), 1)
    assert np.isclose(tt.index_vectorized(0, 2), 2)
    assert np.isclose(tt.index_vectorized(0, 3), 3)
    assert np.isclose(tt.index_vectorized(1, 0), 4)
    assert np.isclose(tt.index_vectorized(1, 1), 5)
    assert np.isclose(tt.index_vectorized(1, 2), 6)
    assert np.isclose(tt.index_vectorized(1, 3), 7)


def test_tensor_iterator():
    a = np.arange(16).reshape((4, 4))
    test_tensor = Tensor(tensor=a)
    assert np.allclose(test_tensor.data, a)
    assert test_tensor.size == 16
    assert isinstance(test_tensor.basis, Bijection)

    a_triu = a[np.triu_indices_from(a)]
    a_tril = a[np.tril_indices_from(a)]

    counter = 0
    for val, idx in test_tensor.utri_iterator():
        assert val == a[tuple(idx)]
        assert val == a_triu[counter]
        counter += 1
    assert counter == 4 * (4 + 1) / 2

    counter = 0
    for val, idx in test_tensor.ltri_iterator():
        assert val == a[tuple(idx)]
        assert val == a_tril[counter]
        counter += 1
    assert counter == 4 * (4 + 1) / 2

    counter = 0
    for val, idx in test_tensor.all_iterator():
        assert val == a[tuple(idx)]
        counter += 1

    assert np.allclose(test_tensor.vectorize(), a.reshape((-1, 1), order='C'))

    with pytest.raises(TypeError):
        list(test_tensor._iterator('blah'))


def test_get_obj_size():
    assert Tensor.get_obj_size(1) == 1
    assert Tensor.get_obj_size([1, 1]) == 2
    with pytest.raises(TypeError):
        Tensor.get_obj_size('a')


def test_index_bijection():
    with pytest.raises(TypeError):
        Tensor.index_bijection((1, 1, 1, 1), 1, 2)



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Tests for lcu_util.py."""

import unittest
import random
import numpy

from openfermion.ops.representations import DiagonalCoulombHamiltonian
from openfermion.transforms import jordan_wigner
from openfermion.circuits.lcu_util import (
    lambda_norm, _discretize_probability_distribution,
    _preprocess_for_efficient_roulette_selection,
    preprocess_lcu_coefficients_for_reversible_sampling)


class LambdaNormTest(unittest.TestCase):

    def test_random(self):

        # Create random DiagonalCoulombHamiltonian.
        n_qubits = 8
        random.seed(n_qubits)
        one_body = numpy.zeros((n_qubits, n_qubits), float)
        two_body = numpy.zeros((n_qubits, n_qubits), float)
        for p in range(8):
            for q in range(p, 8):
                one_body[p, q] = random.random()
                two_body[p, q] = random.random()
        one_body += one_body.T
        two_body += two_body.T
        diagonal_operator = DiagonalCoulombHamiltonian(one_body, two_body,
                                                       random.random())

        # Compute the lambda norm using expensive (reliable) method.
        qubit_operator = jordan_wigner(diagonal_operator)
        del qubit_operator.terms[()]
        reliable_norm = qubit_operator.induced_norm()

        # Compute the lambda norm using the method under test.
        test_norm = lambda_norm(diagonal_operator)

        # Third norm.
        third_norm = 0.
        for _, coefficient in qubit_operator.terms.items():
            third_norm += abs(coefficient)

        # Test.
        self.assertAlmostEqual(third_norm, reliable_norm)
        self.assertAlmostEqual(test_norm, reliable_norm)


class DiscretizeDistributionTest(unittest.TestCase):

    def assertGetDiscretizedDistribution(self, probabilities, epsilon):
        total_probability = sum(probabilities)
        numers, denom, mu = _discretize_probability_distribution(
            probabilities, epsilon)
        self.assertEqual(sum(numers), denom)
        self.assertEqual(len(numers), len(probabilities))
        self.assertEqual(len(probabilities) * 2**mu, denom)
        for i in range(len(numers)):
            self.assertAlmostEqual(numers[i] / denom,
                                   probabilities[i] / total_probability,
                                   delta=epsilon)
        return numers, denom

    def test_fuzz(self):
        random.seed(8)
        for _ in range(100):
            n = random.randint(1, 50)
            weights = [random.random() for _ in range(n)]
            self.assertGetDiscretizedDistribution(weights,
                                                  2**-random.randint(1, 20))

    def test_known_discretizations(self):
        self.assertEqual(self.assertGetDiscretizedDistribution([1], 0.25),
                         ([4], 4))

        self.assertEqual(self.assertGetDiscretizedDistribution([1], 0.125),
                         ([8], 8))

        self.assertEqual(
            self.assertGetDiscretizedDistribution([0.1, 0.1, 0.1], 0.25),
            ([2, 2, 2], 6))

        self.assertEqual(
            self.assertGetDiscretizedDistribution([0.09, 0.11, 0.1], 0.25),
            ([2, 2, 2], 6))

        self.assertEqual(
            self.assertGetDiscretizedDistribution([0.09, 0.11, 0.1], 0.1),
            ([4, 4, 4], 12))

        self.assertEqual(
            self.assertGetDiscretizedDistribution([0.09, 0.11, 0.1], 0.05),
            ([7, 9, 8], 24))

        self.assertEqual(
            self.assertGetDiscretizedDistribution([0.09, 0.11, 0.1], 0.01),
            ([58, 70, 64], 192))

        self.assertEqual(
            self.assertGetDiscretizedDistribution([0.09, 0.11, 0.1], 0.00335),
            ([115, 141, 128], 384))


class PreprocessForEfficientRouletteSelectionTest(unittest.TestCase):

    def assertPreprocess(self, weights):
        alternates, keep_chances = (
            _preprocess_for_efficient_roulette_selection(weights))

        self.assertEqual(len(alternates), len(keep_chances))

        target_weight = sum(weights) // len(alternates)
        distribution = list(keep_chances)
        for i in range(len(alternates)):
            distribution[alternates[i]] += target_weight - keep_chances[i]
        self.assertEqual(weights, distribution)

        return alternates, keep_chances

    def test_fuzz(self):
        random.seed(8)
        for _ in range(100):
            n = random.randint(1, 50)
            weights = [random.randint(0, 100) for _ in range(n)]
            weights[-1] += n - sum(weights) % n  # Ensure multiple of length.
            self.assertPreprocess(weights)

    def test_validation(self):
        with self.assertRaises(ValueError):
            _ = self.assertPreprocess(weights=[])
        with self.assertRaises(ValueError):
            _ = self.assertPreprocess(weights=[1, 2])
        with self.assertRaises(ValueError):
            _ = self.assertPreprocess(weights=[3, 3, 2])

    def test_already_uniform(self):
        self.assertEqual(self.assertPreprocess(weights=[1]), ([0], [0]))
        self.assertEqual(self.assertPreprocess(weights=[1, 1]),
                         ([0, 1], [0, 0]))
        self.assertEqual(self.assertPreprocess(weights=[1, 1, 1]),
                         ([0, 1, 2], [0, 0, 0]))
        self.assertEqual(self.assertPreprocess(weights=[2, 2, 2]),
                         ([0, 1, 2], [0, 0, 0]))

    def test_donation(self):
        # v2 donates 1 to v0.
        self.assertEqual(self.assertPreprocess(weights=[1, 2, 3]),
                         ([2, 1, 2], [1, 0, 0]))
        # v0 donates 1 to v1.
        self.assertEqual(self.assertPreprocess(weights=[3, 1, 2]),
                         ([0, 0, 2], [0, 1, 0]))
        # v0 donates 1 to v1, then 2 to v2.
        self.assertEqual(self.assertPreprocess(weights=[5, 1, 0]),
                         ([0, 0, 0], [0, 1, 0]))

    def test_over_donation(self):
        # v0 donates 2 to v1, leaving v0 needy, then v2 donates 1 to v0.
        self.assertEqual(self.assertPreprocess(weights=[3, 0, 3]),
                         ([2, 0, 2], [1, 0, 0]))


class PreprocessLCUCoefficientsForReversibleSamplingTest(unittest.TestCase):

    def assertPreprocess(self, lcu_coefs, epsilon):
        alternates, keep_numers, mu = (
            preprocess_lcu_coefficients_for_reversible_sampling(
                lcu_coefs, epsilon))

        n = len(lcu_coefs)
        keep_denom = 2**mu
        self.assertEqual(len(alternates), n)
        self.assertEqual(len(keep_numers), n)
        self.assertTrue(all(0 <= e < keep_denom for e in keep_numers))

        out_distribution = [1 / n * numer / keep_denom for numer in keep_numers]
        for i in range(n):
            switch_probability = 1 - keep_numers[i] / keep_denom
            out_distribution[alternates[i]] += 1 / n * switch_probability

        total = sum(lcu_coefs)
        for i in range(n):
            self.assertAlmostEqual(out_distribution[i],
                                   lcu_coefs[i] / total,
                                   delta=epsilon)

        return alternates, keep_numers, keep_denom

    def test_fuzz(self):
        random.seed(8)
        for _ in range(100):
            n = random.randint(1, 50)
            weights = [random.randint(0, 100) for _ in range(n)]
            weights[-1] += n - sum(weights) % n  # Ensure multiple of length.
            self.assertPreprocess(weights, 2**-random.randint(1, 20))

    def test_known(self):
        self.assertEqual(self.assertPreprocess([1, 2], epsilon=0.01),
                         ([1, 1], [43, 0], 64))

        self.assertEqual(self.assertPreprocess([1, 2, 3], epsilon=0.01),
                         ([2, 1, 2], [32, 0, 0], 64))



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""
This module should contain objects that describe the algebra of various
operators and representations which can be thought of as abstract instances that
serve as storage of the operators.  The abstract instance objects derive from
the PolynomialTensor.  Here we differentiate between generic storage objects and
particular instantiations.
"""
from .operators import (
    BinaryPolynomial,
    BinaryPolynomialError,
    BosonOperator,
    FermionOperator,
    IsingOperator,
    MajoranaOperator,
    QuadOperator,
    QubitOperator,
    SymbolicOperator,
    double_decoding,
    shift_decoder,
    BinaryCode,
    BinaryCodeError,
)

from .representations import (
    PolynomialTensor,
    PolynomialTensorError,
    general_basis_change,
    DiagonalCoulombHamiltonian,
    InteractionOperator,
    InteractionOperatorError,
    InteractionRDM,
    InteractionRDMError,
    QuadraticHamiltonian,
    QuadraticHamiltonianError,
    DOCIHamiltonian,
    antisymmetric_canonical_form,
)



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Tests for pubchem.py."""

import unittest
import numpy
import pytest

from openfermion.chem.pubchem import geometry_from_pubchem
from openfermion.testing.testing_utils import module_importable

using_pubchempy = pytest.mark.skipif(module_importable('pubchempy') is False,
                                     reason='Not detecting `pubchempy`.')


@using_pubchempy
class OpenFermionPubChemTest(unittest.TestCase):

    def test_water(self):
        water_geometry = geometry_from_pubchem('water')
        self.water_natoms = len(water_geometry)
        self.water_atoms = [water_atom[0] for water_atom in water_geometry]
        water_oxygen_index = self.water_atoms.index('O')
        water_oxygen = water_geometry.pop(water_oxygen_index)
        water_oxygen_coordinate = numpy.array(water_oxygen[1])
        water_hydrogen1_coordinate = numpy.array(water_geometry[0][1])
        water_hydrogen2_coordinate = numpy.array(water_geometry[1][1])
        water_oxygen_hydrogen1 = \
            water_hydrogen1_coordinate - water_oxygen_coordinate
        water_oxygen_hydrogen2 = \
            water_hydrogen2_coordinate - water_oxygen_coordinate

        self.water_bond_length_1 = numpy.linalg.norm(water_oxygen_hydrogen1)
        self.water_bond_length_2 = numpy.linalg.norm(water_oxygen_hydrogen2)
        self.water_bond_angle = \
            numpy.arccos(numpy.dot(water_oxygen_hydrogen1,
                                   water_oxygen_hydrogen2 /
                                   (numpy.linalg.norm(water_oxygen_hydrogen1) *
                                    numpy.linalg.norm(water_oxygen_hydrogen2))))

        water_natoms = 3
        self.assertEqual(water_natoms, self.water_natoms)

        self.assertAlmostEqual(self.water_bond_length_1,
                               self.water_bond_length_2,
                               places=4)
        water_bond_length_low = 0.9
        water_bond_length_high = 1.1
        self.assertTrue(water_bond_length_low <= self.water_bond_length_1)
        self.assertTrue(water_bond_length_high >= self.water_bond_length_1)

        water_bond_angle_low = 100. / 360 * 2 * numpy.pi
        water_bond_angle_high = 110. / 360 * 2 * numpy.pi
        self.assertTrue(water_bond_angle_low <= self.water_bond_angle)
        self.assertTrue(water_bond_angle_high >= self.water_bond_angle)

    def test_helium(self):
        helium_geometry = geometry_from_pubchem('helium')
        self.helium_natoms = len(helium_geometry)

        helium_natoms = 1
        self.assertEqual(helium_natoms, self.helium_natoms)

    def test_none(self):
        none_geometry = geometry_from_pubchem('none')

        self.assertIsNone(none_geometry)

    def test_water_2d(self):
        water_geometry = geometry_from_pubchem('water', structure='2d')
        self.water_natoms = len(water_geometry)

        water_natoms = 3
        self.assertEqual(water_natoms, self.water_natoms)

        self.oxygen_z_1 = water_geometry[0][1][2]
        self.oxygen_z_2 = water_geometry[1][1][2]
        z = 0
        self.assertEqual(z, self.oxygen_z_1)
        self.assertEqual(z, self.oxygen_z_2)

        with pytest.raises(ValueError,
                           match='Incorrect value for the argument structure'):
            _ = geometry_from_pubchem('water', structure='foo')



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Classes for representing algorithms for performing Trotter steps."""

from typing import Optional, Sequence, TYPE_CHECKING, Tuple, Union

import abc

import cirq
from openfermion import ops

if TYPE_CHECKING:
    # pylint: disable=unused-import
    from typing import Set, Type

Hamiltonian = Union[ops.FermionOperator, ops.QubitOperator, ops.
                    InteractionOperator, ops.DiagonalCoulombHamiltonian]


class TrotterStep(metaclass=abc.ABCMeta):
    """A method for performing a Trotter step.

    This class assumes that Hamiltonian evolution using a Trotter-Suzuki product
    formula is performed in the following steps:

        1. Perform some preparatory operations (for instance, a basis change).
        2. Perform a number of Trotter steps. Each Trotter step may induce a
           permutation on the ordering in which qubits represent fermionic
           modes.
        3. Perform some finishing operations.

    Attributes:
        hamiltonian: The Hamiltonian being simulated.
    """

    def __init__(self, hamiltonian: Hamiltonian) -> None:
        self.hamiltonian = hamiltonian

    def prepare(self,
                qubits: Sequence[cirq.Qid],
                control_qubit: Optional[cirq.Qid] = None) -> cirq.OP_TREE:
        """Operations to perform before doing the Trotter steps.

        Args:
            qubits: The qubits on which to perform operations. They should
                be sorted so that the j-th qubit in the Sequence holds the
                occupation of the j-th fermionic mode.
            hamiltonian: The Hamiltonian to simulate.
            control_qubit: The control qubit, if the algorithm is controlled.
        """
        # Default: do nothing
        return ()

    @abc.abstractmethod
    def trotter_step(self,
                     qubits: Sequence[cirq.Qid],
                     time: float,
                     control_qubit: Optional[cirq.Qid] = None) -> cirq.OP_TREE:
        """Yield operations to perform a Trotter step.

        Args:
            qubits: The qubits on which to apply the Trotter step.
            hamiltonian: The Hamiltonian to simulate.
            time: The evolution time.
            control_qubit: The control qubit, if the algorithm is controlled.
        """

    def step_qubit_permutation(
            self,
            qubits: Sequence[cirq.Qid],
            control_qubit: Optional[cirq.Qid] = None
    ) -> Tuple[Sequence[cirq.Qid], Optional[cirq.Qid]]:
        """The qubit permutation induced by a single Trotter step.

        Returns:
            A tuple whose first element is the new list of system qubits and
            second element is the new control qubit
        """
        # Default: identity permutation
        return qubits, control_qubit

    def finish(self,
               qubits: Sequence[cirq.Qid],
               n_steps: int,
               control_qubit: Optional[cirq.Qid] = None,
               omit_final_swaps: bool = False) -> cirq.OP_TREE:
        """Operations to perform after all Trotter steps are done.

        Args:
            qubits: The qubits on which to perform operations.
            hamiltonian: The Hamiltonian to simulate.
            n_steps: The total number of Trotter steps that have been performed.
            control_qubit: The control qubit, if the algorithm is controlled.
            omit_final_swaps: Whether or not to omit swap gates at the end of
                the circuit.
        """
        # Default: do nothing
        return ()


class TrotterAlgorithm(metaclass=abc.ABCMeta):
    """An algorithm for performing a Trotter step.

    A Trotter step algorithm contains methods for performing a symmetric or
    asymmetric Trotter step and their controlled versions. It does not need
    to support all the possibilities; for instance, it may support only
    symmetric Trotter steps with no control qubit. Support for a kind of
    Trotter step is implemented by overriding the methods of this class.

    Attributes:
        supported_types: A set containing types of Hamiltonian representations
            that can be simulated using this Trotter step algorithm.
            For example, {DiagonalCoulombHamiltonian, InteractionOperator}.
    """
    supported_types = set()  # type: Set[Type[Hamiltonian]]

    def symmetric(self, hamiltonian: Hamiltonian) -> Optional[TrotterStep]:
        return None

    def asymmetric(self, hamiltonian: Hamiltonian) -> Optional[TrotterStep]:
        return None

    def controlled_symmetric(self,
                             hamiltonian: Hamiltonian) -> Optional[TrotterStep]:
        return None

    def controlled_asymmetric(self, hamiltonian: Hamiltonian
                             ) -> Optional[TrotterStep]:
        return None



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from .bch_expansion import bch_expand

from .channel_state import (
    amplitude_damping_channel,
    dephasing_channel,
    depolarizing_channel,
)

# Imports out of alphabetical order to avoid circular dependency.
from .lattice import (
    HubbardSquareLattice,
    HubbardLattice,
    SpinPairs,
    Spin,
)

from .commutators import (
    anticommutator,
    commutator,
    double_commutator,
    trivially_double_commutes_dual_basis_using_term_info,
    trivially_commutes_dual_basis,
    trivially_double_commutes_dual_basis,
)

from .grid import Grid

from .indexing import (
    up_index,
    down_index,
    up_then_down,
)

from .operator_utils import (
    count_qubits,
    get_file_path,
    hermitian_conjugated,
    is_hermitian,
    is_identity,
    load_operator,
    save_operator,
    OperatorUtilsError,
    OperatorSpecificationError,
)

from .rdm_mapping_functions import (
    kronecker_delta,
    map_two_pdm_to_two_hole_dm,
    map_two_pdm_to_one_pdm,
    map_one_pdm_to_one_hole_dm,
    map_one_hole_dm_to_one_pdm,
    map_two_pdm_to_particle_hole_dm,
    map_two_hole_dm_to_two_pdm,
    map_two_hole_dm_to_one_hole_dm,
    map_particle_hole_dm_to_one_pdm,
    map_particle_hole_dm_to_two_pdm,
)



import os
import numpy
from openfermion.config import DATA_DIRECTORY
from openfermion.chem import MolecularData
from openfermion.ops.representations import InteractionOperator
from openfermion.chem.reduced_hamiltonian import make_reduced_hamiltonian
from openfermion.linalg.sparse_tools import (
    get_number_preserving_sparse_operator)
from openfermion.transforms.opconversions import get_fermion_operator


def test_mrd_return_type():
    filename = os.path.join(DATA_DIRECTORY, "H2_sto-3g_singlet_0.7414.hdf5")
    molecule = MolecularData(filename=filename)
    reduced_ham = make_reduced_hamiltonian(molecule.get_molecular_hamiltonian(),
                                           molecule.n_electrons)

    assert isinstance(reduced_ham, InteractionOperator)


def test_constant_one_body():
    filename = os.path.join(DATA_DIRECTORY, "H2_sto-3g_singlet_0.7414.hdf5")
    molecule = MolecularData(filename=filename)
    reduced_ham = make_reduced_hamiltonian(molecule.get_molecular_hamiltonian(),
                                           molecule.n_electrons)

    assert numpy.isclose(reduced_ham.constant, molecule.nuclear_repulsion)
    assert numpy.allclose(reduced_ham.one_body_tensor, 0)


def test_fci_energy():
    filename = os.path.join(DATA_DIRECTORY, "H2_sto-3g_singlet_0.7414.hdf5")
    molecule = MolecularData(filename=filename)
    reduced_ham = make_reduced_hamiltonian(molecule.get_molecular_hamiltonian(),
                                           molecule.n_electrons)
    numpy_ham = get_number_preserving_sparse_operator(
        get_fermion_operator(reduced_ham),
        molecule.n_qubits,
        num_electrons=molecule.n_electrons,
        spin_preserving=True)

    w, _ = numpy.linalg.eigh(numpy_ham.toarray())
    assert numpy.isclose(molecule.fci_energy, w[0])

    filename = os.path.join(DATA_DIRECTORY, "H1-Li1_sto-3g_singlet_1.45.hdf5")
    molecule = MolecularData(filename=filename)
    reduced_ham = make_reduced_hamiltonian(molecule.get_molecular_hamiltonian(),
                                           molecule.n_electrons)
    numpy_ham = get_number_preserving_sparse_operator(
        get_fermion_operator(reduced_ham),
        molecule.n_qubits,
        num_electrons=molecule.n_electrons,
        spin_preserving=True)

    w, _ = numpy.linalg.eigh(numpy_ham.toarray())
    assert numpy.isclose(molecule.fci_energy, w[0])



from .polynomial_tensor import (
    PolynomialTensor,
    PolynomialTensorError,
    general_basis_change,
)

from .diagonal_coulomb_hamiltonian import DiagonalCoulombHamiltonian

from .interaction_operator import (
    InteractionOperator,
    InteractionOperatorError,
    get_tensors_from_integrals,
    get_active_space_integrals,
)

from .interaction_rdm import (
    InteractionRDM,
    InteractionRDMError,
)

from .quadratic_hamiltonian import (
    QuadraticHamiltonian,
    QuadraticHamiltonianError,
    antisymmetric_canonical_form,
)
from .doci_hamiltonian import DOCIHamiltonian



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
""" Code to generate Pauli strings for measurement of local operators"""
from itertools import chain, zip_longest
import numpy

from openfermion.ops.operators import QubitOperator


def binary_partition_iterator(qubit_list, num_iterations=None):
    """Generator for a list of 2-partitions of N qubits
    such that all pairs of qubits are split in at least one partition,
    This follows a variation on ArXiv:1908.0562 - instead of
    explicitly partitioning the list based on the binary indices of
    the qubits, we repeatedly divide the list in two and then
    zip it back together.

    Args:
        qubit_list(list): list of qubits to be partitioned
        num_iterations(int or None): number of iterations to perform.
            If None, will be set to ceil(log2(len(qubit_list)))

    Returns:
        partition(iterator of tuples of lists): the required partitioning
    """

    # Some edge cases
    if num_iterations is not None and num_iterations == 0:
        return
    num_qubits = len(qubit_list)
    if num_qubits < 2:
        raise ValueError('Need at least 2 qubits to partition')
    if num_qubits == 2:
        yield ([qubit_list[0]], [qubit_list[1]])
        return

    if num_iterations is None:
        num_iterations = int(numpy.ceil(numpy.log2(num_qubits)))

    # Calculate the point where we need to split the list each time.
    half_point = int(numpy.ceil(num_qubits / 2))

    # Repeat the division and zip steps as many times
    # as required.
    for _ in range(num_iterations):
        # Divide the qubit list in two and return it
        partition = (qubit_list[:half_point], qubit_list[half_point:])
        yield partition
        # Zip the partition together to remake the qubit list.
        qubit_list = list(chain(*zip_longest(partition[0], partition[1])))
        # If len(qubit_list) is odd, the end of the list will be 'None'
        # which we delete.
        if qubit_list[-1] is None:
            del qubit_list[-1]


def partition_iterator(qubit_list, partition_size, num_iterations=None):
    """Generator for a list of k-partitions of N qubits such that
    all sets of k qubits are perfectly split in at least one
    partition, following ArXiv:1908.05628

    Args:
        qubit_list(list): list of qubits to be partitioned
        partition_size(int): the number of sets in the partition.
        num_iterations(int or None): the number of iterations in the
            outer iterator. If None, set to ceil(log2(len(qubit_list)))

    Returns:
        partition(iterator of tuples of lists): the required partitioning
    """

    # Some edge cases
    if num_iterations == 0:
        return
    if partition_size == 1:
        yield (qubit_list,)
        return
    elif partition_size == 2:
        for p in binary_partition_iterator(qubit_list, num_iterations):
            yield p
        return
    num_qubits = len(qubit_list)
    if partition_size == num_qubits:
        yield tuple([q] for q in qubit_list)
        return
    elif partition_size > num_qubits:
        raise ValueError('I cant k-partition less than k qubits')

    if num_iterations is None:
        num_iterations = int(numpy.ceil(numpy.log2(num_qubits)))

    # First iterate over the outer binary partition
    outer_iterator = binary_partition_iterator(qubit_list,
                                               num_iterations=num_iterations)
    for set1, set2 in outer_iterator:

        # Each new partition needs to be subdivided fewer times
        # to prevent an additional k! factor in the scaling.
        num_iterations -= 1

        # Iterate over all possibilities of partitioning the first
        # set into l parts and the second set into k - l parts.
        for inner_partition_size in range(1, partition_size):
            if inner_partition_size > len(set1) or\
                    partition_size - inner_partition_size > len(set2):
                continue

            # subdivide the first partition
            inner_iterator1 = partition_iterator(set1, inner_partition_size,
                                                 num_iterations)
            for inner_partition1 in inner_iterator1:

                # subdivide the second partition
                inner_iterator2 = partition_iterator(
                    set2, partition_size - inner_partition_size, num_iterations)
                for inner_partition2 in inner_iterator2:
                    yield inner_partition1 + inner_partition2


def pauli_string_iterator(num_qubits, max_word_size=2):
    """Generates a set of Pauli strings such that each word
    of k Pauli operators lies in at least one string.

    Args:
        num_qubits(int): number of qubits in string
        max_word_size(int): maximum required word

    Returns:
        pauli_string(iterator of strings): iterator
            over Pauli strings
    """
    if max_word_size > num_qubits:
        raise ValueError('Number of qubits is too few')
    if max_word_size <= 0:
        raise ValueError('Word size too small')

    qubit_list = list(range(num_qubits))
    partitions = partition_iterator(qubit_list, max_word_size)
    pauli_string = ['I' for temp in range(num_qubits)]
    pauli_letters = ['X', 'Y', 'Z']
    for partition in partitions:
        for lettering in range(3**max_word_size):
            for p in partition:
                letter = pauli_letters[lettering % 3]
                for qubit in p:
                    pauli_string[qubit] = letter
                lettering = lettering // 3
            yield tuple(pauli_string)


def _find_compatible_basis(term, bases):
    for basis in bases:
        basis_qubits = {op[0] for op in basis}
        conflicts = ((i, P)
                     for (i, P) in term
                     if i in basis_qubits and (i, P) not in basis)
        if any(conflicts):
            continue
        return basis
    return None


def group_into_tensor_product_basis_sets(operator, seed=None):
    """
    Split an operator (instance of QubitOperator) into `sub-operator`
    QubitOperators, where each sub-operator has terms that are diagonal
    in the same tensor product basis.

    Each `sub-operator` can be measured using the same qubit post-rotations
    in expectation estimation. Grouping into these tensor product basis
    sets has been found to improve the efficiency of expectation estimation
    significantly for some Hamiltonians in the context of
    VQE (see section V(A) in the supplementary material of
    https://arxiv.org/pdf/1704.05018v2.pdf). The more general problem
    of grouping operators into commutitative groups is discussed in
    section IV (B2) of https://arxiv.org/pdf/1509.04279v1.pdf. The
    original input operator is the union of all output sub-operators,
    and all sub-operators are disjoint (do not share any terms).

    Args:
        operator (QubitOperator): the operator that will be split into
            sub-operators (tensor product basis sets).
        seed (int): default None. Random seed used to initialize the
            numpy.RandomState pseudo-random number generator.

    Returns:
        sub_operators (dict): a dictionary where each key defines a
            tensor product basis, and each corresponding value is a
            QubitOperator with terms that are all diagonal in
            that basis.
            **key** (tuple of tuples): Each key is a term, which defines
                a tensor product basis. A term is a product of individual
                factors; each factor is represented by a tuple of the form
                (`index`, `action`), and these tuples are collected into a
                larger tuple which represents the term as the product of
                its factors. `action` is from the set {'X', 'Y', 'Z'} and
                `index` is a non-negative integer corresponding to the
                index of a qubit.
            **value** (QubitOperator): A QubitOperator with terms that are
                diagonal in the basis defined by the key it is stored in.

    Raises:
       TypeError: Operator of invalid type.
    """
    if not isinstance(operator, QubitOperator):
        raise TypeError('Can only split QubitOperator into tensor product'
                        ' basis sets. {} is not supported.'.format(
                            type(operator).__name__))

    sub_operators = {}
    r = numpy.random.RandomState(seed)
    for term, coefficient in operator.terms.items():
        bases = list(sub_operators.keys())
        r.shuffle(bases)
        basis = _find_compatible_basis(term, bases)
        if basis is None:
            sub_operators[term] = QubitOperator(term, coefficient)
        else:
            sub_operator = sub_operators.pop(basis)
            sub_operator += QubitOperator(term, coefficient)
            additions = tuple(op for op in term if op not in basis)
            basis = tuple(
                sorted(basis + additions, key=lambda factor: factor[0]))
            sub_operators[basis] = sub_operator

    return sub_operators



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import numpy
import pytest
import cirq
from cirq import LineQubit

import openfermion
from openfermion.linalg import get_sparse_operator
from openfermion.circuits.primitives.state_preparation import (
    prepare_slater_determinant, prepare_gaussian_state)
from openfermion.testing import random_quadratic_hamiltonian


@pytest.mark.parametrize(
    'n_qubits, conserves_particle_number, occupied_orbitals, initial_state',
    [(4, True, None, 0b0010), (4, False, None, 0b1001), (5, True, None, 0),
     (5, False, None, 0b10101), (5, True, range(4), 0),
     (5, False, (0, 2, 3), [1, 2, 3, 4])])
def test_prepare_gaussian_state(n_qubits,
                                conserves_particle_number,
                                occupied_orbitals,
                                initial_state,
                                atol=1e-5):

    qubits = LineQubit.range(n_qubits)

    # Initialize a random quadratic Hamiltonian
    quad_ham = random_quadratic_hamiltonian(n_qubits,
                                            conserves_particle_number,
                                            real=False)
    quad_ham_sparse = get_sparse_operator(quad_ham)

    # Compute the energy of the desired state
    if occupied_orbitals is None:
        energy = quad_ham.ground_energy()
    else:
        orbital_energies, _, constant = (
            quad_ham.diagonalizing_bogoliubov_transform())
        energy = sum(orbital_energies[i] for i in occupied_orbitals) + constant

    # Get the state using a circuit simulation
    circuit = cirq.Circuit(
        prepare_gaussian_state(qubits,
                               quad_ham,
                               occupied_orbitals,
                               initial_state=initial_state))
    if isinstance(initial_state, list):
        initial_state = sum(1 << (n_qubits - 1 - i) for i in initial_state)
    state = circuit.final_state_vector(initial_state=initial_state)

    # Check that the result is an eigenstate with the correct eigenvalue
    numpy.testing.assert_allclose(quad_ham_sparse.dot(state),
                                  energy * state,
                                  atol=atol)


@pytest.mark.parametrize(
    'n_spatial_orbitals, conserves_particle_number, occupied_orbitals, '
    'initial_state', [(4, True, [range(1), range(1)], 0b00100100),
                      (5, True, [range(2), range(1)], list(range(3))),
                      (5, True, [[0, 2], [1, 3]], 0)])
def test_prepare_gaussian_state_with_spin_symmetry(n_spatial_orbitals,
                                                   conserves_particle_number,
                                                   occupied_orbitals,
                                                   initial_state,
                                                   atol=1e-5):

    n_qubits = 2 * n_spatial_orbitals
    qubits = LineQubit.range(n_qubits)

    # Initialize a random quadratic Hamiltonian
    quad_ham = random_quadratic_hamiltonian(n_spatial_orbitals,
                                            conserves_particle_number,
                                            real=True,
                                            expand_spin=True,
                                            seed=639)

    # Reorder the Hamiltonian and get sparse matrix
    quad_ham = openfermion.get_quadratic_hamiltonian(
        openfermion.reorder(openfermion.get_fermion_operator(quad_ham),
                            openfermion.up_then_down))
    quad_ham_sparse = get_sparse_operator(quad_ham)

    # Compute the energy of the desired state
    energy = 0.0
    for spin_sector in range(2):
        orbital_energies, _, _ = (quad_ham.diagonalizing_bogoliubov_transform(
            spin_sector=spin_sector))
        energy += sum(
            orbital_energies[i] for i in occupied_orbitals[spin_sector])
    energy += quad_ham.constant

    # Get the state using a circuit simulation
    circuit = cirq.Circuit(
        prepare_gaussian_state(qubits,
                               quad_ham,
                               occupied_orbitals,
                               initial_state=initial_state))

    if isinstance(initial_state, list):
        initial_state = sum(1 << (n_qubits - 1 - i) for i in initial_state)
    state = circuit.final_state_vector(initial_state=initial_state)

    # Check that the result is an eigenstate with the correct eigenvalue
    numpy.testing.assert_allclose(quad_ham_sparse.dot(state),
                                  energy * state,
                                  atol=atol)


@pytest.mark.parametrize(
    'slater_determinant_matrix, correct_state, initial_state', [
        (numpy.array([[1, 1]]) / numpy.sqrt(2),
         numpy.array([0, 1, 1, 0]) / numpy.sqrt(2), 0),
        (numpy.array([[1, 1j]]) / numpy.sqrt(2),
         numpy.array([0, 1j, 1, 0]) / numpy.sqrt(2), 0b01),
        (numpy.array(
            [[1, 1, 1],
             [1, numpy.exp(2j * numpy.pi / 3),
              numpy.exp(4j * numpy.pi / 3)]]) / numpy.sqrt(3),
         numpy.array([
             0, 0, 0,
             numpy.exp(2j * numpy.pi / 3), 0, 1 + numpy.exp(2j * numpy.pi / 3),
             1, 0
         ]) / numpy.sqrt(3), 0),
        (numpy.array(
            [[1, 1, 1],
             [1, numpy.exp(2j * numpy.pi / 3),
              numpy.exp(4j * numpy.pi / 3)]]) / numpy.sqrt(3),
         numpy.array([
             0, 0, 0,
             numpy.exp(2j * numpy.pi / 3), 0, 1 + numpy.exp(2j * numpy.pi / 3),
             1, 0
         ]) / numpy.sqrt(3), [0, 2]),
    ])
def test_prepare_slater_determinant(slater_determinant_matrix,
                                    correct_state,
                                    initial_state,
                                    atol=1e-7):

    n_qubits = slater_determinant_matrix.shape[1]
    qubits = LineQubit.range(n_qubits)

    circuit = cirq.Circuit(
        prepare_slater_determinant(qubits,
                                   slater_determinant_matrix,
                                   initial_state=initial_state))
    if isinstance(initial_state, list):
        initial_state = sum(1 << (n_qubits - 1 - i) for i in initial_state)
    state = circuit.final_state_vector(initial_state=initial_state)

    assert cirq.allclose_up_to_global_phase(state, correct_state, atol=atol)



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from .algorithms import (
    LINEAR_SWAP_NETWORK,
    LinearSwapNetworkTrotterAlgorithm,
    LOW_RANK,
    LowRankTrotterAlgorithm,
    SplitOperatorTrotterAlgorithm,
    SplitOperatorTrotterStep,
    SymmetricSplitOperatorTrotterStep,
    ControlledAsymmetricSplitOperatorTrotterStep,
    AsymmetricSplitOperatorTrotterStep,
    ControlledSymmetricSplitOperatorTrotterStep,
    SPLIT_OPERATOR,
)

from .diagonal_coulomb_trotter_error import (
    diagonal_coulomb_potential_and_kinetic_terms_as_arrays,
    bit_mask_of_modes_acted_on_by_fermionic_terms,
    split_operator_trotter_error_operator_diagonal_two_body,
    fermionic_swap_trotter_error_operator_diagonal_two_body,
)

from .hubbard_trotter_error import \
    simulation_ordered_grouped_hubbard_terms_with_info

from .low_depth_trotter_error import (
    low_depth_second_order_trotter_error_operator,
    low_depth_second_order_trotter_error_bound,
    simulation_ordered_grouped_low_depth_terms_with_info,
    stagger_with_info,
)

from .simulate_trotter import simulate_trotter

from .trotter_algorithm import TrotterAlgorithm, TrotterStep

from .trotter_error import (
    error_bound,
    error_operator,
    trotter_steps_required,
)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from .circuit_validation import validate_trotterized_evolution

from .random import random_interaction_operator_term

from .testing_utils import (
    haar_random_vector,
    random_antisymmetric_matrix,
    random_hermitian_matrix,
    random_unitary_matrix,
    random_qubit_operator,
    random_diagonal_coulomb_hamiltonian,
    random_interaction_operator,
    random_quadratic_hamiltonian,
    EqualsTester,
    module_importable,
)

from .wrapped import (
    assert_equivalent_repr,
    assert_implements_consistent_protocols,
)



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import pytest
import numpy

from openfermion.ops.operators import MajoranaOperator


def test_majorana_operator_init():
    op = MajoranaOperator((0, 7, 4, 1, 10))
    assert op.terms == {(0, 1, 4, 7, 10): -1.0}

    op = MajoranaOperator((0, 1, 0, 1, 0))
    assert op.terms == {(0,): -1.0}

    op = MajoranaOperator((3, 2, 5, 1, 3, 4))
    assert op.terms == {(1, 2, 4, 5): 1.0}

    op = MajoranaOperator((5, 10, 4, 3, 6, 9, 6))
    assert op.terms == {(3, 4, 5, 9, 10): -1.0}


def test_majorana_operator_commutes_with():
    a = MajoranaOperator((0, 1, 5))
    b = MajoranaOperator((1, 2, 7))
    c = MajoranaOperator((2, 3, 4))
    d = MajoranaOperator((0, 3, 6))

    assert a.commutes_with(b)
    assert not a.commutes_with(c)
    assert a.commutes_with(d)
    assert b.commutes_with(c)
    assert not b.commutes_with(d)
    assert c.commutes_with(d)
    assert (a + c).commutes_with(b + d)

    e = MajoranaOperator((0, 1, 1, 1, 4, 5))
    f = MajoranaOperator((0, 1, 1, 4))

    assert e.commutes_with(f)

    with pytest.raises(TypeError):
        _ = e.commutes_with(0)


def test_majorana_operator_with_basis_rotated_by():
    H = numpy.array([[1, 1], [1, -1]]) / numpy.sqrt(2)

    a = MajoranaOperator((0, 1), 2.0)
    op = a.with_basis_rotated_by(H)
    assert op == MajoranaOperator.from_dict({(0, 1): -2.0})

    b = MajoranaOperator((0,), 2.0)
    op = b.with_basis_rotated_by(H)
    assert op == MajoranaOperator.from_dict({
        (0,): numpy.sqrt(2),
        (1,): numpy.sqrt(2)
    })

    c = MajoranaOperator((1,), 2.0)
    op = c.with_basis_rotated_by(H)
    assert op == MajoranaOperator.from_dict({
        (0,): numpy.sqrt(2),
        (1,): -numpy.sqrt(2)
    })

    P = numpy.array([[0, 1, 0], [0, 0, 1], [1, 0, 0]])

    d = MajoranaOperator((0, 1, 2)) + MajoranaOperator((1, 2))
    op = d.with_basis_rotated_by(P)
    assert op == MajoranaOperator.from_dict({(0, 1, 2): 1.0, (0, 1): 1.0})

    with pytest.raises(ValueError):
        _ = a.with_basis_rotated_by(2 * H)


def test_majorana_operator_add_subtract():
    a = MajoranaOperator((0, 2, 3), -1.25)
    b = MajoranaOperator((0, 2, 3), -0.5)
    c = MajoranaOperator((1, 5, 7), 4.75)
    d = MajoranaOperator((3, 5, 7), 2.25)

    a += c
    assert a.terms == {(0, 2, 3): -1.25, (1, 5, 7): 4.75}

    a -= d
    assert a.terms == {(0, 2, 3): -1.25, (1, 5, 7): 4.75, (3, 5, 7): 2.25}

    a += MajoranaOperator((0, 2, 3), 0.5)
    assert a.terms == {(0, 2, 3): -.75, (1, 5, 7): 4.75, (3, 5, 7): 2.25}

    a -= MajoranaOperator((0, 2, 3), 0.25)
    assert a.terms == {(0, 2, 3): -1.0, (1, 5, 7): 4.75, (3, 5, 7): 2.25}

    assert (a + b).terms == {(0, 2, 3): -1.5, (1, 5, 7): 4.75, (3, 5, 7): 2.25}

    assert (a - b).terms == {(0, 2, 3): -0.5, (1, 5, 7): 4.75, (3, 5, 7): 2.25}

    with pytest.raises(TypeError):
        _ = a + 0

    with pytest.raises(TypeError):
        a += 0

    with pytest.raises(TypeError):
        _ = a - 0

    with pytest.raises(TypeError):
        a -= 0


def test_majorana_operator_multiply():
    a = MajoranaOperator((0, 1, 5), 1.5) + MajoranaOperator((1, 2, 7), -0.5)
    b = MajoranaOperator((2, 3, 4), 1.75) - MajoranaOperator((0, 3, 6), 0.25)
    assert (a * a).terms == {(): -2.5, (0, 2, 5, 7): -1.5}
    assert (a * b).terms == {
        (1, 3, 4, 7): 0.875,
        (0, 1, 2, 3, 6, 7): -0.125,
        (1, 3, 5, 6): 0.375,
        (0, 1, 2, 3, 4, 5): -2.625
    }
    assert (2 * a).terms == (a * 2).terms == {(0, 1, 5): 3.0, (1, 2, 7): -1.0}

    a *= 2
    a *= MajoranaOperator(())
    assert a.terms == {(0, 1, 5): 3.0, (1, 2, 7): -1.0}

    with pytest.raises(TypeError):
        _ = a * 'a'

    with pytest.raises(TypeError):
        _ = 'a' * a

    with pytest.raises(TypeError):
        a *= 'a'


def test_majorana_operator_pow():
    a = MajoranaOperator((0, 1, 5), 1.5) + MajoranaOperator((1, 2, 7), -0.5)
    assert (a**2).terms == {(): -2.5, (0, 2, 5, 7): -1.5}

    with pytest.raises(TypeError):
        _ = a**-1

    with pytest.raises(TypeError):
        _ = a**'a'


def test_majorana_operator_divide():
    a = MajoranaOperator((0, 1, 5), 1.5) + MajoranaOperator((1, 2, 7), -0.5)
    assert (a / 2).terms == {(0, 1, 5): 0.75, (1, 2, 7): -0.25}

    a /= 2
    assert a.terms == {(0, 1, 5): 0.75, (1, 2, 7): -0.25}

    with pytest.raises(TypeError):
        _ = a / 'a'

    with pytest.raises(TypeError):
        a /= 'a'


def test_majorana_operator_neg():
    a = MajoranaOperator((0, 1, 5), 1.5) + MajoranaOperator((1, 2, 7), -0.5)
    assert (-a).terms == {(0, 1, 5): -1.5, (1, 2, 7): 0.5}


def test_majorana_operator_eq():
    a = MajoranaOperator((0, 1, 5), 1.5) + MajoranaOperator((1, 2, 7), -0.5)
    b = (MajoranaOperator((0, 1, 5), 1.5) + MajoranaOperator(
        (1, 2, 7), -0.5) + MajoranaOperator((3, 4, 5), 0.0))
    c = (MajoranaOperator((0, 1, 5), 1.5) + MajoranaOperator(
        (1, 2, 7), -0.5) + MajoranaOperator((3, 4, 5), 0.1))
    d = MajoranaOperator((0, 1, 5), 1.75) + MajoranaOperator((1, 2, 7), -0.75)
    e = MajoranaOperator((0, 1, 5), 1.5) - MajoranaOperator((0, 3, 6), 0.25)

    assert a == b
    assert a != c
    assert a != d
    assert a != e

    assert a != 0


def test_majorana_operator_str():
    zero = MajoranaOperator()
    one = MajoranaOperator(())
    still_zero = MajoranaOperator((0, 1, 5), 0.0)
    a = MajoranaOperator((0, 1, 5), 1.5)
    b = MajoranaOperator((1, 2, 7), -0.5)

    assert str(zero) == '0'
    assert str(one) == '1.0 ()'
    assert str(still_zero) == '0'
    assert str(a) == '1.5 (0, 1, 5)'
    assert str(b) == '-0.5 (1, 2, 7)'
    assert str(a + b) == """1.5 (0, 1, 5) +
-0.5 (1, 2, 7)"""


def test_majorana_operator_repr():
    a = MajoranaOperator((0, 1, 5), 1.5)
    assert repr(a) == 'MajoranaOperator.from_dict(terms={(0, 1, 5): 1.5})'



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Class and functions to store reduced density matrices."""
import copy
import numpy

from openfermion.ops.operators import FermionOperator, QubitOperator
from openfermion.ops.representations import (InteractionOperator,
                                             PolynomialTensor)


class InteractionRDMError(Exception):
    pass


class InteractionRDM(PolynomialTensor):
    r"""Class for storing 1- and 2-body reduced density matrices.

    Attributes:
        one_body_tensor: The expectation values <a^\dagger_p a_q>.
        two_body_tensor: The expectation values
            <a^\dagger_p a^\dagger_q a_r a_s>.
    """

    def __init__(self, one_body_tensor, two_body_tensor):
        r"""Initialize the InteractionRDM class.

        Args:
            one_body_tensor: Expectation values <a^\dagger_p a_q>.
            two_body_tensor: Expectation values
                <a^\dagger_p a^\dagger_q a_r a_s>.
        """
        super(InteractionRDM, self).__init__({
            (1, 0): one_body_tensor,
            (1, 1, 0, 0): two_body_tensor
        })

    @property
    def one_body_tensor(self):
        """The value of the one-body tensor."""
        return self.n_body_tensors[1, 0]

    @one_body_tensor.setter
    def one_body_tensor(self, value):
        """Set the value of the one-body tensor."""
        self.n_body_tensors[1, 0] = value

    @property
    def two_body_tensor(self):
        """The value of the two-body tensor."""
        return self.n_body_tensors[1, 1, 0, 0]

    @two_body_tensor.setter
    def two_body_tensor(self, value):
        """Set the value of the two-body tensor."""
        self.n_body_tensors[1, 1, 0, 0] = value

    def expectation(self, operator):
        """Return expectation value of an InteractionRDM with an operator.

        Args:
            operator: A QubitOperator or InteractionOperator.

        Returns:
            float: Expectation value

        Raises:
            InteractionRDMError: Invalid operator provided.
        """
        if isinstance(operator, QubitOperator):
            expectation_op = self.get_qubit_expectations(operator)
            expectation = 0.0
            for qubit_term in operator.terms:
                expectation += (operator.terms[qubit_term] *
                                expectation_op.terms[qubit_term])
        elif isinstance(operator, InteractionOperator):
            expectation = operator.constant
            expectation += numpy.sum(self.one_body_tensor *
                                     operator.one_body_tensor)
            expectation += numpy.sum(self.two_body_tensor *
                                     operator.two_body_tensor)
        else:
            raise InteractionRDMError('Invalid operator type provided.')
        return expectation

    def get_qubit_expectations(self, qubit_operator):
        """Return expectations of QubitOperator in new QubitOperator.

        Args:
            qubit_operator: QubitOperator instance to be evaluated on
                this InteractionRDM.

        Returns:
            QubitOperator: QubitOperator with coefficients
            corresponding to expectation values of those operators.

        Raises:
            InteractionRDMError: Observable not contained in 1-RDM or 2-RDM.
        """
        # Importing here instead of head of file to prevent circulars
        from openfermion.transforms.opconversions import (reverse_jordan_wigner,
                                                          normal_ordered)
        qubit_operator_expectations = copy.deepcopy(qubit_operator)
        for qubit_term in qubit_operator_expectations.terms:
            expectation = 0.

            # Map qubits back to fermions.
            reversed_fermion_operators = reverse_jordan_wigner(
                QubitOperator(qubit_term))
            reversed_fermion_operators = normal_ordered(
                reversed_fermion_operators)

            # Loop through fermion terms.
            for fermion_term in reversed_fermion_operators.terms:
                coefficient = reversed_fermion_operators.terms[fermion_term]

                # Handle molecular term.
                if FermionOperator(
                        fermion_term).is_two_body_number_conserving():
                    if not fermion_term:
                        expectation += coefficient
                    else:
                        indices = [operator[0] for operator in fermion_term]
                        if len(indices) == 2:
                            # One-body term
                            indices = tuple(zip(indices, (1, 0)))
                        else:
                            # Two-body term
                            indices = tuple(zip(indices, (1, 1, 0, 0)))
                        rdm_element = self[indices]
                        expectation += rdm_element * coefficient

                # Handle non-molecular terms.
                elif len(fermion_term) > 4:
                    raise InteractionRDMError('Observable not contained '
                                              'in 1-RDM or 2-RDM.')
            qubit_operator_expectations.terms[qubit_term] = expectation
        return qubit_operator_expectations



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""BosonOperator stores a sum of products of bosonic ladder operators."""

from openfermion.ops.operators.symbolic_operator import SymbolicOperator


class BosonOperator(SymbolicOperator):
    r"""BosonOperator stores a sum of products of bosonic ladder operators.

    In OpenFermion, we describe bosonic ladder operators using the shorthand:
    'i^' = b^\dagger_i
    'j' = b_j
    where ['i', 'j^'] = delta_ij is the commutator.

    One can multiply together these bosonic ladder operators to obtain a
    bosonic term. For instance, '2^ 1' is a bosonic term which
    creates at mode 2 and destroys at mode 1. The BosonicOperator class
    also stores a coefficient for the term, e.g. '3.17 * 2^ 1'.

    The BosonOperator class is designed (in general) to store sums of these
    terms. For instance, an instance of BosonOperator might represent
    3.17 2^ 1 - 66.2 * 8^ 7 6^ 2
    The Bosonic Operator class overloads operations for manipulation of
    these objects by the user.

    BosonOperator is a subclass of SymbolicOperator. Importantly, it has
    attributes set as follows::

        actions = (1, 0)
        action_strings = ('^', '')
        action_before_index = False
        different_indices_commute = True

    See the documentation of SymbolicOperator for more details.

    Example:
        .. code-block:: python

            H = (BosonOperator('0^ 3', .5)
                   + .5 * BosonOperator('3^ 0'))
            # Equivalently
            H2 = BosonOperator('0^ 3', 0.5)
            H2 += BosonOperator('3^ 0', 0.5)

    Note:
        Adding BosonOperator is faster using += (as this
        is done by in-place addition). Specifying the coefficient
        during initialization is faster than multiplying a BosonOperator
        with a scalar.
    """

    @property
    def actions(self):
        """The allowed actions."""
        return (1, 0)

    @property
    def action_strings(self):
        """The string representations of the allowed actions."""
        return ('^', '')

    @property
    def action_before_index(self):
        """Whether action comes before index in string representations."""
        return False

    @property
    def different_indices_commute(self):
        """Whether factors acting on different indices commute."""
        return True

    def is_normal_ordered(self):
        """Return whether or not term is in normal order.

        In our convention, ladder operators come first.
        Note that unlike the Fermion operator, due to the commutation
        of ladder operators with different indices, the BosonOperator
        sorts ladder operators by index.
        """
        for term in self.terms:
            for i in range(1, len(term)):
                for j in range(i, 0, -1):
                    right_operator = term[j]
                    left_operator = term[j - 1]
                    if (right_operator[0] == left_operator[0] and
                            right_operator[1] > left_operator[1]):
                        return False
        return True

    def is_boson_preserving(self):
        """Query whether the term preserves particle number.

        This is equivalent to requiring the same number of
        raising and lowering operators in each term.
        """
        for term in self.terms:
            # Make sure term conserves particle number
            particles = 0
            for operator in term:
                particles += (-1)**operator[1]  # add 1 if create, else -1
            if not (particles == 0):
                return False
        return True



# Out of order to fix circular import
from .term_reordering import (
    chemist_ordered,
    normal_ordered,
    normal_ordered_ladder_term,
    normal_ordered_quad_term,
    reorder,
)

from .binary_codes import (
    linearize_decoder,
    checksum_code,
    bravyi_kitaev_code,
    jordan_wigner_code,
    parity_code,
    weight_one_binary_addressing_code,
    weight_one_segment_code,
    weight_two_segment_code,
    interleaved_code,
)

from .binary_code_transform import (
    binary_code_transform,
    extractor,
    dissolve,
    make_parity_list,
)

from .bksf import (
    bravyi_kitaev_fast,
    bravyi_kitaev_fast_interaction_op,
    bravyi_kitaev_fast_edge_matrix,
    number_operator,
    vacuum_operator,
    edge_operator_aij,
    edge_operator_b,
    generate_fermions,
)

from .bravyi_kitaev import (
    bravyi_kitaev,
    inline_sum,
    inline_product,
)

from .bravyi_kitaev_tree import bravyi_kitaev_tree

from .commutator_diagonal_coulomb_operator import (
    commutator_ordered_diagonal_coulomb_with_two_body_operator,)

from .conversions import (
    get_fermion_operator,
    get_boson_operator,
    get_majorana_operator,
    get_quad_operator,
    check_no_sympy,
)

from .fenwick_tree import (
    FenwickNode,
    FenwickTree,
)

from .jordan_wigner import (
    jordan_wigner,
    jordan_wigner_one_body,
    jordan_wigner_two_body,
)

from .qubitoperator_to_paulisum import qubit_operator_to_pauli_sum

from .reverse_jordan_wigner import reverse_jordan_wigner

from .remove_symmetry_qubits import (
    symmetry_conserving_bravyi_kitaev,
    edit_hamiltonian_for_spin,
)

from .verstraete_cirac import (
    verstraete_cirac_2d_square,
    vertical_edges_snake,
)



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import abc
from enum import Enum, IntEnum
import itertools


class Spin(IntEnum):
    UP = 0
    DOWN = 1


class SpinPairs(Enum):
    """The spin orbitals corresponding to a pair of spatial orbitals."""

    SAME = 1
    ALL = 0
    DIFF = -1


class HubbardLattice(metaclass=abc.ABCMeta):
    """Base class for a Hubbard model lattice.

    Subclasses must define the following properties:
        n_dofs (int): The number of degrees of freedom per site (and spin if
            applicable).
        n_sites (int): The number of sites in the lattice.
        spinless (bool): Whether or not the fermion has spin (False if so).
        edge_types (Tuple[Hashable, ...]): The types of edges that a term could
            correspond to. Examples include 'onsite', 'neighbor',
            'diagonal_neighbor', etc.
        onsite_edge_types (Sequence[Hashable]): The edge types that connect
            sites to themselves.

    And the following methods:
        site_pairs_iter(edge_type: Hashable) \\
            -> Iterable[Tuple[int, int]]: Iterable
            over pairs of sites corresponding to the given edge type.

    For 'spinful' lattices, the ``spin_indices`` ``0`` and ``1`` correspond to
    'up' and 'down', respectively.
    """

    @abc.abstractproperty
    def n_dofs(self):
        """The number of degrees of freedom per site
        (and spin if applicable)."""

    @abc.abstractproperty
    def n_sites(self):
        """The number of sites in the lattice."""

    @abc.abstractproperty
    def spinless(self):
        """Whether or not the fermion has spin (False if so)."""

    @abc.abstractproperty
    def edge_types(self):
        """The types of edges that a term could correspond to.

        Examples include 'onsite', 'neighbor', 'diagonal_neighbor', etc.
        """

    @abc.abstractproperty
    def onsite_edge_types(self):
        """The edge types that connect sites to themselves."""

    @abc.abstractmethod
    def site_pairs_iter(self, edge_type, ordered=True):
        """Iterable over pairs of sites corresponding to the given edge type."""

    # properties

    @property
    def n_spin_values(self):
        return 1 if self.spinless else 2

    @property
    def n_spin_orbitals_per_site(self):
        return self.n_dofs * self.n_spin_values

    @property
    def n_spin_orbitals(self):
        return self.n_sites * self.n_spin_orbitals_per_site

    # indexing

    def site_index_offset(self, site_index):
        return site_index * self.n_spin_orbitals_per_site

    def dof_index_offset(self, dof_index):
        return dof_index * self.n_spin_values

    def to_spin_orbital_index(self, site_index, dof_index, spin_index):
        """The index of the spin orbital."""
        return (self.site_index_offset(site_index) +
                self.dof_index_offset(dof_index) + spin_index)

    def from_spin_orbital_index(self, spin_orbital_index):
        site_index, offset = divmod(spin_orbital_index,
                                    self.n_spin_orbitals_per_site)
        dof_index, spin_index = divmod(offset, self.n_spin_values)
        return site_index, dof_index, spin_index

    # iteration

    @property
    def site_indices(self):
        return range(self.n_sites)

    @property
    def dof_indices(self):
        return range(self.n_dofs)

    def dof_pairs_iter(self, exclude_same=False):
        return ((a, b)
                for a in range(self.n_dofs)
                for b in range(a + exclude_same, self.n_dofs))

    @property
    def spin_indices(self):
        return range(self.n_spin_values)

    def spin_pairs_iter(self, spin_pairs=SpinPairs.ALL, ordered=True):
        if spin_pairs == SpinPairs.ALL:
            return (itertools.product(self.spin_indices, repeat=2)
                    if ordered else itertools.combinations_with_replacement(
                        self.spin_indices, 2))
        elif spin_pairs == SpinPairs.SAME:
            return ((s, s) for s in self.spin_indices)
        elif spin_pairs == SpinPairs.DIFF:
            return (itertools.permutations(self.spin_indices, 2) if ordered else
                    itertools.combinations(self.spin_indices, 2))
        raise ValueError(
            '{} not a valid SpinPairs specification.'.format(spin_pairs))

    # validation

    def validate_edge_type(self, edge_type):
        if edge_type not in self.edge_types:
            raise ValueError('{} not a valid edge type {}.'.format(
                edge_type, self.edge_types))

    def validate_dof(self, dof, length=None):
        if not (0 <= dof < self.n_dofs):
            raise ValueError('not (0 <= {} < n_dofs = {})'.format(
                dof, self.n_dofs))

    def validate_dofs(self, dofs, length=None):
        for dof in dofs:
            self.validate_dof(dof)
        if (length is not None) and (len(dofs) != length):
            raise ValueError('len({}) != {}'.format(dofs, length))


class HubbardSquareLattice(HubbardLattice):
    r"""A square lattice for a Hubbard model.

    Valid edge types are:
        * 'onsite'
        * 'horizontal_neighbor'
        * 'vertical_neighbor'
        * 'neighbor': union of 'horizontal_neighbor' and 'vertical_neighbor'
        * 'diagonal_neighbor'
    """

    def __init__(self,
                 x_dimension,
                 y_dimension,
                 n_dofs=1,
                 spinless=False,
                 periodic=True):
        """
        Args:
            x_dimension (int): The width of the grid.
            y_dimension (int): The height of the grid.
            n_dofs (int, optional): The number of degrees of freedom per site
                (and spin if applicable). Defaults is 1.
            periodic (bool, optional): If True, add periodic boundary
                conditions. Default is True.
            spinless (bool, optional): If True, return a spinless Fermi-Hubbard
                model. Default is False.
        """

        self.x_dimension = x_dimension
        self.y_dimension = y_dimension
        self._n_dofs = n_dofs
        self._spinless = spinless
        self.periodic = periodic

    @property
    def n_dofs(self):
        return self._n_dofs

    @property
    def spinless(self):
        return self._spinless

    @property
    def n_sites(self):
        return self.x_dimension * self.y_dimension

    @property
    def edge_types(self):
        return ('onsite', 'neighbor', 'diagonal_neighbor',
                'horizontal_neighbor', 'vertical_neighbor')

    @property
    def onsite_edge_types(self):
        return ('onsite',)

    def site_pairs_iter(self, edge_type, ordered=True):
        if edge_type == 'onsite':
            return ((i, i) for i in self.site_indices)
        elif edge_type == 'neighbor':
            return self.neighbors_iter(ordered)
        elif edge_type == 'horizontal_neighbor':
            return self.horizontal_neighbors_iter(ordered)
        elif edge_type == 'vertical_neighbor':
            return self.vertical_neighbors_iter(ordered)
        elif edge_type == 'diagonal_neighbor':
            return self.diagonal_neighbors_iter(ordered)
        raise ValueError('Edge type {} is not valid.'.format(edge_type))

    def __repr__(self):
        return '{}({})'.format(
            self.__class__.__name__, ', '.join(
                (('x_dimension={}'.format(self.x_dimension)),
                 ('y_dimension={}'.format(self.y_dimension)),
                 ('n_dofs={}'.format(self.n_dofs)), ('spinless={}'.format(
                     self.spinless)), ('periodic={}'.format(self.periodic)))))

    # site indexing

    def to_site_index(self, site):
        """The index of a site."""
        x, y = site
        return x + y * self.x_dimension

    def from_site_index(self, site_index):
        return divmod(site_index, self.x_dimension)[::-1]

    # neighbor counting and iteration

    def n_horizontal_neighbor_pairs(self, ordered=True):
        """Number of horizontally neighboring (unordered) pairs of sites."""
        n_horizontal_edges_per_y = (
            self.x_dimension - (self.x_dimension <= 2 or not self.periodic))
        return (self.y_dimension * n_horizontal_edges_per_y *
                (2 if ordered else 1))

    def n_vertical_neighbor_pairs(self, ordered=True):
        """Number of vertically neighboring (unordered) pairs of sites."""
        n_vertical_edges_per_x = (self.y_dimension -
                                  (self.y_dimension <= 2 or not self.periodic))
        return (self.x_dimension * n_vertical_edges_per_x *
                (2 if ordered else 1))

    def n_neighbor_pairs(self, ordered=True):
        """Number of neighboring (unordered) pairs of sites."""
        return (self.n_horizontal_neighbor_pairs(ordered) +
                self.n_vertical_neighbor_pairs(ordered))

    def neighbors_iter(self, ordered=True):
        return itertools.chain(self.horizontal_neighbors_iter(ordered),
                               self.vertical_neighbors_iter(ordered))

    def diagonal_neighbors_iter(self, ordered=True):
        n_sites_per_y = (self.x_dimension -
                         (self.x_dimension <= 2 or not self.periodic))
        n_sites_per_x = (self.y_dimension -
                         (self.y_dimension <= 2 or not self.periodic))
        for x in range(n_sites_per_y):
            for y in range(n_sites_per_x):
                for dy in (-1, 1):
                    i = self.to_site_index((x, y))
                    j = self.to_site_index(((x + 1) % self.x_dimension,
                                            (y + dy) % self.y_dimension))
                    yield (i, j)
                    if ordered:
                        yield (j, i)

    def horizontal_neighbors_iter(self, ordered=True):
        n_horizontal_edges_per_y = (
            self.x_dimension - (self.x_dimension <= 2 or not self.periodic))
        for x in range(n_horizontal_edges_per_y):
            for y in range(self.y_dimension):
                i = self.to_site_index((x, y))
                j = self.to_site_index(((x + 1) % self.x_dimension, y))
                yield (i, j)
                if ordered:
                    yield (j, i)

    def vertical_neighbors_iter(self, ordered=True):
        n_vertical_edges_per_x = (self.y_dimension -
                                  (self.y_dimension <= 2 or not self.periodic))
        for y in range(n_vertical_edges_per_x):
            for x in range(self.x_dimension):
                i = self.to_site_index((x, y))
                j = self.to_site_index((x, (y + 1) % self.y_dimension))
                yield (i, j)
                if ordered:
                    yield (j, i)

    # square-specific geometry

    @property
    def shape(self):
        return (self.x_dimension, self.y_dimension)

    def delta_mag(self, X, Y, by_index=False):
        """The distance between sites X and Y in each dimension."""
        if by_index:
            return self.delta_mag(self.from_site_index(X),
                                  self.from_site_index(Y))
        if self.periodic:
            return tuple(
                min(abs((s * (x - y)) % d)
                    for s in (-1, 1))
                for d, x, y in zip(self.shape, X, Y))
        return tuple(abs(x - xx) for x, xx in zip(X, Y))

    def manhattan_distance(self, X, Y, by_index=False):
        return sum(self.delta_mag(X, Y, by_index))



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Tests for BosonOperator."""

import unittest

from openfermion.ops.operators.boson_operator import BosonOperator
from openfermion.hamiltonians import number_operator


class BosonOperatorTest(unittest.TestCase):

    def test_is_normal_ordered_empty(self):
        op = BosonOperator() * 2
        self.assertTrue(op.is_normal_ordered())

    def test_is_normal_ordered_number(self):
        op = BosonOperator('2^ 2') * -1j
        self.assertTrue(op.is_normal_ordered())

    def test_is_normal_ordered_reversed(self):
        self.assertFalse(BosonOperator('2 2^').is_normal_ordered())

    def test_is_normal_ordered_create(self):
        self.assertTrue(BosonOperator('11^').is_normal_ordered())

    def test_is_normal_ordered_annihilate(self):
        self.assertTrue(BosonOperator('0').is_normal_ordered())

    def test_is_normal_ordered_long_not(self):
        self.assertTrue(BosonOperator('0 5^ 3^ 2^ 1^').is_normal_ordered())

    def test_is_normal_ordered_outoforder(self):
        self.assertTrue(BosonOperator('0 1').is_normal_ordered())

    def test_is_normal_ordered_long_descending(self):
        self.assertTrue(BosonOperator('5^ 3^ 2^ 1^ 0').is_normal_ordered())

    def test_is_normal_ordered_multi(self):
        op = BosonOperator('4 3 2^ 2') + BosonOperator('1 2')
        self.assertTrue(op.is_normal_ordered())

    def test_is_normal_ordered_multiorder(self):
        op = BosonOperator('4 3 2 1') + BosonOperator('3 2')
        self.assertTrue(op.is_normal_ordered())

    def test_is_boson_preserving_BosonOperator(self):
        op = BosonOperator()
        self.assertTrue(op.is_boson_preserving())

    def test_is_boson_preserving_number(self):
        op = number_operator(n_modes=5, mode=3, parity=1)
        self.assertTrue(op.is_boson_preserving())

    def test_is_boson_preserving_three(self):
        op = BosonOperator(((0, 1), (2, 1), (4, 0)))
        self.assertFalse(op.is_boson_preserving())

    def test_is_boson_preserving_out_of_order(self):
        op = BosonOperator(((0, 1), (2, 0), (1, 1), (3, 0)))
        self.assertTrue(op.is_boson_preserving())



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""QuadOperator stores a sum of products of canonical quadrature operators."""

from openfermion.ops.operators.symbolic_operator import SymbolicOperator


class QuadOperator(SymbolicOperator):
    """QuadOperator stores a sum of products of canonical quadrature operators.

    They are defined in terms of the bosonic ladder operators:
    q = sqrt{hbar/2}(b+b^)
    p = -isqrt{hbar/2}(b-b^)
    where hbar is a constant appearing in the commutator of q and p:
    [q, p] = i hbar

    In OpenFermion, we describe the canonical quadrature operators acting
    on quantum modes 'i' and 'j' using the shorthand:
    'qi' = q_i
    'pj' = p_j
    where ['qi', 'pj'] = i hbar delta_ij is the commutator.

    The QuadOperator class is designed (in general) to store sums of these
    terms. For instance, an instance of QuadOperator might represent

    .. code-block:: python

        H = 0.5 * QuadOperator('q0 p5') + 0.3 * QuadOperator('q0')

    Note for a QuadOperator to be a Hamiltonian which is a hermitian
    operator, the coefficients of all terms must be real.

    QuadOperator is a subclass of SymbolicOperator. Importantly, it has
    attributes set as follows::

        actions = ('q', 'p')
        action_strings = ('q', 'p')
        action_before_index = True
        different_indices_commute = True

    See the documentation of SymbolicOperator for more details.

    Example:
        .. code-block:: python

            H = (QuadOperator('p0 q3', 0.5)
                   + 0.6 * QuadOperator('p3 q0'))
            # Equivalently
            H2 = QuadOperator('p0 q3', 0.5)
            H2 += QuadOperator('p3 q0', 0.6)

    Note:
        Adding QuadOperator is faster using += (as this
        is done by in-place addition). Specifying the coefficient
        during initialization is faster than multiplying a QuadOperator
        with a scalar.
    """

    @property
    def actions(self):
        """The allowed actions."""
        return ('q', 'p')

    @property
    def action_strings(self):
        """The string representations of the allowed actions."""
        return ('q', 'p')

    @property
    def action_before_index(self):
        """Whether action comes before index in string representations."""
        return True

    @property
    def different_indices_commute(self):
        """Whether factors acting on different indices commute."""
        return True

    def is_normal_ordered(self):
        """Return whether or not term is in normal order.

        In our convention, q operators come first.
        Note that unlike the Fermion operator, due to the commutation
        of quadrature operators with different indices, the QuadOperator
        sorts quadrature operators by index.
        """
        for term in self.terms:
            for i in range(1, len(term)):
                for j in range(i, 0, -1):
                    right_operator = term[j]
                    left_operator = term[j - 1]
                    if (right_operator[0] == left_operator[0] and
                            right_operator[1] == 'q' and
                            left_operator[1] == 'p'):
                        return False
        return True

    def is_gaussian(self):
        """Query whether the term is quadratic or lower in the
        quadrature operators.
        """
        for term in self.terms:
            if len(term) > 2:
                return False

        return True



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Common gates that target three qubits."""
import numpy as np
import cirq


def rot111(rads: float) -> cirq.CCZPowGate:
    """Phases the |111> state of three qubits by e^{i rads}."""
    return cirq.CCZ**(rads / np.pi)


def CRxxyy(rads: float) -> cirq.ControlledGate:
    """Controlled version of openfermion.Rxxyy"""
    return cirq.ControlledGate(cirq.ISwapPowGate(exponent=-2 * rads / np.pi))


def CRyxxy(rads: float) -> cirq.ControlledGate:
    """Controlled version of openfermion.Ryxxy"""
    return cirq.ControlledGate(
        cirq.PhasedISwapPowGate(exponent=2 * rads / np.pi))



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import numpy

from openfermion.config import EQ_TOLERANCE
from openfermion.ops.operators import FermionOperator
from openfermion.ops.representations import (DiagonalCoulombHamiltonian,
                                             InteractionOperator,
                                             InteractionOperatorError)
from openfermion.transforms.opconversions import (check_no_sympy,
                                                  normal_ordered)
from openfermion.ops.representations.quadratic_hamiltonian import (
    QuadraticHamiltonian, QuadraticHamiltonianError)

from openfermion.chem import MolecularData

# for breaking cyclic imports
from openfermion.utils import operator_utils as op_utils


def get_quadratic_hamiltonian(fermion_operator,
                              chemical_potential=0.,
                              n_qubits=None,
                              ignore_incompatible_terms=False):
    r"""Convert a quadratic fermionic operator to QuadraticHamiltonian.

    Args:
        fermion_operator(FermionOperator): The operator to convert.
        chemical_potential(float): A chemical potential to include in
            the returned operator
        n_qubits(int): Optionally specify the total number of qubits in the
            system
        ignore_incompatible_terms(bool): This flag determines the behavior
            of this method when it encounters terms that are not quadratic
            that is, terms that are not of the form a^\dagger_p a_q.
            If set to True, this method will simply ignore those terms.
            If False, then this method will raise an error if it encounters
            such a term. The default setting is False.

    Returns:
       quadratic_hamiltonian: An instance of the QuadraticHamiltonian class.

    Raises:
        TypeError: Input must be a FermionOperator.
        TypeError: FermionOperator does not map to QuadraticHamiltonian.

    Warning:
        Even assuming that each creation or annihilation operator appears
        at most a constant number of times in the original operator, the
        runtime of this method is exponential in the number of qubits.
    """
    if not isinstance(fermion_operator, FermionOperator):
        raise TypeError('Input must be a FermionOperator.')

    check_no_sympy(fermion_operator)

    if n_qubits is None:
        n_qubits = op_utils.count_qubits(fermion_operator)
    if n_qubits < op_utils.count_qubits(fermion_operator):
        raise ValueError('Invalid number of qubits specified.')

    # Normal order the terms and initialize.
    fermion_operator = normal_ordered(fermion_operator)
    constant = 0.
    combined_hermitian_part = numpy.zeros((n_qubits, n_qubits), complex)
    antisymmetric_part = numpy.zeros((n_qubits, n_qubits), complex)

    # Loop through terms and assign to matrix.
    for term in fermion_operator.terms:
        coefficient = fermion_operator.terms[term]
        # Ignore this term if the coefficient is zero
        if abs(coefficient) < EQ_TOLERANCE:
            # not testable because normal_ordered kills
            # fermion terms lower than EQ_TOLERANCE
            continue  # pragma: no cover

        if len(term) == 0:
            # Constant term
            constant = coefficient
        elif len(term) == 2:
            ladder_type = [operator[1] for operator in term]
            p, q = [operator[0] for operator in term]

            if ladder_type == [1, 0]:
                combined_hermitian_part[p, q] = coefficient
            elif ladder_type == [1, 1]:
                # Need to check that the corresponding [0, 0] term is present
                conjugate_term = ((p, 0), (q, 0))
                if conjugate_term not in fermion_operator.terms:
                    raise QuadraticHamiltonianError(
                        'FermionOperator does not map '
                        'to QuadraticHamiltonian (not Hermitian).')
                else:
                    matching_coefficient = -fermion_operator.terms[
                        conjugate_term].conjugate()
                    discrepancy = abs(coefficient - matching_coefficient)
                    if discrepancy > EQ_TOLERANCE:
                        raise QuadraticHamiltonianError(
                            'FermionOperator does not map '
                            'to QuadraticHamiltonian (not Hermitian).')
                antisymmetric_part[p, q] += .5 * coefficient
                antisymmetric_part[q, p] -= .5 * coefficient
            else:
                # ladder_type == [0, 0]
                # Need to check that the corresponding [1, 1] term is present
                conjugate_term = ((p, 1), (q, 1))
                if conjugate_term not in fermion_operator.terms:
                    raise QuadraticHamiltonianError(
                        'FermionOperator does not map '
                        'to QuadraticHamiltonian (not Hermitian).')
                else:
                    matching_coefficient = -fermion_operator.terms[
                        conjugate_term].conjugate()
                    discrepancy = abs(coefficient - matching_coefficient)
                    if discrepancy > EQ_TOLERANCE:
                        raise QuadraticHamiltonianError(
                            'FermionOperator does not map '
                            'to QuadraticHamiltonian (not Hermitian).')
                antisymmetric_part[p, q] -= .5 * coefficient.conjugate()
                antisymmetric_part[q, p] += .5 * coefficient.conjugate()
        elif not ignore_incompatible_terms:
            # Operator contains non-quadratic terms
            raise QuadraticHamiltonianError('FermionOperator does not map '
                                            'to QuadraticHamiltonian '
                                            '(contains non-quadratic terms).')

    # Compute Hermitian part
    hermitian_part = (combined_hermitian_part +
                      chemical_potential * numpy.eye(n_qubits))

    # Check that the operator is Hermitian
    if not op_utils.is_hermitian(hermitian_part):
        raise QuadraticHamiltonianError(
            'FermionOperator does not map '
            'to QuadraticHamiltonian (not Hermitian).')

    # Form QuadraticHamiltonian and return.
    discrepancy = numpy.max(numpy.abs(antisymmetric_part))
    if discrepancy < EQ_TOLERANCE:
        # Hamiltonian conserves particle number
        quadratic_hamiltonian = QuadraticHamiltonian(
            hermitian_part,
            constant=constant,
            chemical_potential=chemical_potential)
    else:
        # Hamiltonian does not conserve particle number
        quadratic_hamiltonian = QuadraticHamiltonian(hermitian_part,
                                                     antisymmetric_part,
                                                     constant,
                                                     chemical_potential)

    return quadratic_hamiltonian


def get_diagonal_coulomb_hamiltonian(fermion_operator,
                                     n_qubits=None,
                                     ignore_incompatible_terms=False):
    r"""Convert a FermionOperator to a DiagonalCoulombHamiltonian.

    Args:
        fermion_operator(FermionOperator): The operator to convert.
        n_qubits(int): Optionally specify the total number of qubits in the
            system
        ignore_incompatible_terms(bool): This flag determines the behavior
            of this method when it encounters terms that are not represented
            by the DiagonalCoulombHamiltonian class, namely, terms that are
            not quadratic and not quartic of the form
            a^\dagger_p a_p a^\dagger_q a_q. If set to True, this method will
            simply ignore those terms. If False, then this method will raise
            an error if it encounters such a term. The default setting is False.
    """
    if not isinstance(fermion_operator, FermionOperator):
        raise TypeError('Input must be a FermionOperator.')

    check_no_sympy(fermion_operator)

    if n_qubits is None:
        n_qubits = op_utils.count_qubits(fermion_operator)
    if n_qubits < op_utils.count_qubits(fermion_operator):
        raise ValueError('Invalid number of qubits specified.')

    fermion_operator = normal_ordered(fermion_operator)
    constant = 0.
    one_body = numpy.zeros((n_qubits, n_qubits), complex)
    two_body = numpy.zeros((n_qubits, n_qubits), float)

    for term, coefficient in fermion_operator.terms.items():
        # Ignore this term if the coefficient is zero
        if abs(coefficient) < EQ_TOLERANCE:
            # not testable because normal_ordered kills
            # fermion terms lower than EQ_TOLERANCE
            continue  # pragma: no cover

        if len(term) == 0:
            constant = coefficient
        else:
            actions = [operator[1] for operator in term]
            if actions == [1, 0]:
                p, q = [operator[0] for operator in term]
                one_body[p, q] = coefficient
            elif actions == [1, 1, 0, 0]:
                p, q, r, s = [operator[0] for operator in term]
                if p == r and q == s:
                    if abs(numpy.imag(coefficient)) > EQ_TOLERANCE:
                        raise ValueError(
                            'FermionOperator does not map to '
                            'DiagonalCoulombHamiltonian (not Hermitian).')
                    coefficient = numpy.real(coefficient)
                    two_body[p, q] = -.5 * coefficient
                    two_body[q, p] = -.5 * coefficient
                elif not ignore_incompatible_terms:
                    raise ValueError('FermionOperator does not map to '
                                     'DiagonalCoulombHamiltonian '
                                     '(contains terms with indices '
                                     '{}).'.format((p, q, r, s)))
            elif not ignore_incompatible_terms:
                raise ValueError('FermionOperator does not map to '
                                 'DiagonalCoulombHamiltonian (contains terms '
                                 'with action {}.'.format(tuple(actions)))

    # Check that the operator is Hermitian
    if not op_utils.is_hermitian(one_body):
        raise ValueError(
            'FermionOperator does not map to DiagonalCoulombHamiltonian '
            '(not Hermitian).')

    return DiagonalCoulombHamiltonian(one_body, two_body, constant)


def get_interaction_operator(fermion_operator, n_qubits=None):
    r"""Convert a 2-body fermionic operator to InteractionOperator.

    This function should only be called on fermionic operators which
    consist of only a_p^\dagger a_q and a_p^\dagger a_q^\dagger a_r a_s
    terms. The one-body terms are stored in a matrix, one_body[p, q], and
    the two-body terms are stored in a tensor, two_body[p, q, r, s].

    Returns:
       interaction_operator: An instance of the InteractionOperator class.

    Raises:
        TypeError: Input must be a FermionOperator.
        TypeError: FermionOperator does not map to InteractionOperator.

    Warning:
        Even assuming that each creation or annihilation operator appears
        at most a constant number of times in the original operator, the
        runtime of this method is exponential in the number of qubits.
    """
    if not isinstance(fermion_operator, FermionOperator):
        raise TypeError('Input must be a FermionOperator.')

    check_no_sympy(fermion_operator)

    if n_qubits is None:
        n_qubits = op_utils.count_qubits(fermion_operator)
    if n_qubits < op_utils.count_qubits(fermion_operator):
        raise ValueError('Invalid number of qubits specified.')

    # Normal order the terms and initialize.
    fermion_operator = normal_ordered(fermion_operator)
    constant = 0.
    one_body = numpy.zeros((n_qubits, n_qubits), complex)
    two_body = numpy.zeros((n_qubits, n_qubits, n_qubits, n_qubits), complex)

    # Loop through terms and assign to matrix.
    for term in fermion_operator.terms:
        coefficient = fermion_operator.terms[term]
        # Ignore this term if the coefficient is zero
        if abs(coefficient) < EQ_TOLERANCE:
            # not testable because normal_ordered kills
            # fermion terms lower than EQ_TOLERANCE
            continue  # pragma: no cover

        # Handle constant shift.
        if len(term) == 0:
            constant = coefficient

        elif len(term) == 2:
            # Handle one-body terms.
            if [operator[1] for operator in term] == [1, 0]:
                p, q = [operator[0] for operator in term]
                one_body[p, q] = coefficient
            else:
                raise InteractionOperatorError('FermionOperator does not map '
                                               'to InteractionOperator.')

        elif len(term) == 4:
            # Handle two-body terms.
            if [operator[1] for operator in term] == [1, 1, 0, 0]:
                p, q, r, s = [operator[0] for operator in term]
                two_body[p, q, r, s] = coefficient
            else:
                raise InteractionOperatorError('FermionOperator does not map '
                                               'to InteractionOperator.')

        else:
            # Handle non-molecular Hamiltonian.
            raise InteractionOperatorError('FermionOperator does not map '
                                           'to InteractionOperator.')

    # Form InteractionOperator and return.
    interaction_operator = InteractionOperator(constant, one_body, two_body)
    return interaction_operator


def get_molecular_data(interaction_operator,
                       geometry=None,
                       basis=None,
                       multiplicity=None,
                       n_electrons=None,
                       reduce_spin=True,
                       data_directory=None):
    """Output a MolecularData object generated from an InteractionOperator

    Args:
        interaction_operator(InteractionOperator): two-body interaction
            operator defining the "molecular interaction" to be simulated.
        geometry(string or list of atoms):
        basis(string):  String denoting the basis set used to discretize the
            system.
        multiplicity(int): Spin multiplicity desired in the system.
        n_electrons(int): Number of electrons in the system
        reduce_spin(bool): True if one wishes to perform spin reduction on
            integrals that are given in interaction operator.  Assumes
            spatial (x) spin structure generically.

    Returns:
        molecule(MolecularData):
            Instance that captures the
            interaction_operator converted into the format that would come
            from an electronic structure package adorned with some meta-data
            that may be useful.
    """

    n_spin_orbitals = interaction_operator.n_qubits

    # Introduce bare molecular operator to fill
    molecule = MolecularData(geometry=geometry,
                             basis=basis,
                             multiplicity=multiplicity,
                             data_directory=data_directory)

    molecule.nuclear_repulsion = interaction_operator.constant

    # Remove spin from integrals and put into molecular operator
    if reduce_spin:
        reduction_indices = list(range(0, n_spin_orbitals, 2))
    else:
        reduction_indices = list(range(n_spin_orbitals))

    molecule.n_orbitals = len(reduction_indices)

    molecule.one_body_integrals = interaction_operator.one_body_tensor[
        numpy.ix_(reduction_indices, reduction_indices)]
    molecule.two_body_integrals = interaction_operator.two_body_tensor[
        numpy.ix_(reduction_indices, reduction_indices, reduction_indices,
                  reduction_indices)]

    # Fill in other metadata
    molecule.overlap_integrals = numpy.eye(molecule.n_orbitals)
    molecule.n_qubits = n_spin_orbitals
    molecule.n_electrons = n_electrons
    molecule.multiplicity = multiplicity

    return molecule



"""
Module performs gradient based RHF, [WIP] UHF, [WIP] GHF

Module needs AO integrals
"""
# pylint: disable=C
from typing import Callable, Dict, List, Optional, Tuple, Union
from itertools import product
import numpy as np
import scipy as sp
from scipy.optimize.optimize import OptimizeResult
from openfermion.ops.representations import (InteractionOperator,
                                             InteractionRDM,
                                             general_basis_change)
import openfermion.linalg as linalg


def get_matrix_of_eigs(w: np.ndarray) -> np.ndarray:
    """
    Transform the eigenvalues into a matrix corresponding
    to summing the adjoint rep.

    Args:
        w: eigenvalues of C-matrix

    Returns: new array of transformed eigenvalues
    """
    transform_eigs = np.zeros((w.shape[0], w.shape[0]), dtype=np.complex128)
    for i, j in product(range(w.shape[0]), repeat=2):
        if np.isclose(abs(w[i] - w[j]), 0):
            transform_eigs[i, j] = 1
        else:
            transform_eigs[i, j] = (np.exp(1j *
                                           (w[i] - w[j])) - 1) / (1j *
                                                                  (w[i] - w[j]))
    return transform_eigs


class InputError(Exception):
    pass


class HartreeFockFunctional():
    """
    Implementation of the objective function code for Restricted Hartree-Fock

    The object transforms a variety of input types into the appropriate output.
    It does this by analyzing the type and size of the input based on its
    knowledge of each type.
    """

    def __init__(self,
                 *,
                 one_body_integrals: np.ndarray,
                 two_body_integrals: np.ndarray,
                 overlap: np.ndarray,
                 n_electrons: int,
                 model='rhf',
                 nuclear_repulsion: Optional[float] = 0.,
                 initial_orbitals: Optional[Union[None, Callable]] = None):
        """
        Initialize functional

        Args:
            one_body_integrals: integrals in the atomic orbital basis for the
                                one-body potential.
            two_body_integrals: integrals in the  atomic obrital basis for the
                                two-body potential ordered according to
                                phi_{p}(r1)^{*}phi_{q}^{*}(r2) x
                                phi_{r}(r2)phi_{s}(r1)
            overlap:  overlap integrals in the atomic orbital basis
            n_electrons:  number of electrons total
            model: Optional flag for performing restricted-, unrestricted-,
                   or generalized- hartree-fock.
            nuclear_repulsion: Optional nuclear repulsion term.  Energy is
                               shifted by this amount. default is 0.
            initial_orbitals:  Method for producing the initial orbitals from
                               the atomic orbitals. Default is defining the
                               core orbitals.
        """
        if model not in ['rhf', 'uhf', 'ghf']:
            raise InputError("{} is not rhf, uhf, or ghf".format(model))
        self.model = model
        self.obi = one_body_integrals
        self.tbi = two_body_integrals
        self.overlap = overlap
        self.num_orbitals = one_body_integrals.shape[0]
        self.num_electrons = n_electrons
        self.constant_offset = nuclear_repulsion
        self.hamiltonian = None

        self.nocc = None
        self.nvirt = None
        self.occ = None
        self.virt = None
        if model == 'rhf':
            self.nocc = self.num_electrons // 2
            self.nvirt = self.num_orbitals - self.nocc
            self.occ = list(range(self.nocc))
            self.virt = list(range(self.nocc, self.nocc + self.nvirt))
        elif model == 'uhf' or model == 'ghf':
            self.nocc = self.num_electrons
            self.nvirt = 2 * self.num_orbitals - self.nocc
            self.occ = list(range(self.nocc))
            self.virt = list(range(self.nocc, self.nocc + self.nvirt))

        if initial_orbitals is None:
            # use core orbitals
            _, core_orbs = sp.linalg.eigh(one_body_integrals, b=overlap)

            molecular_hamiltonian = generate_hamiltonian(
                one_body_integrals=general_basis_change(self.obi, core_orbs,
                                                        (1, 0)),
                two_body_integrals=general_basis_change(self.tbi, core_orbs,
                                                        (1, 1, 0, 0)),
                constant=self.constant_offset)
            self.hamiltonian = molecular_hamiltonian
        else:
            self.hamiltonian = initial_orbitals(self.obi, self.tbi,
                                                self.num_electrons)

    def rdms_from_rhf_opdm(self, opdm_aa: np.ndarray) -> InteractionRDM:
        """
        Generate spin-orbital InteractionRDM object from the alpha-spin
        opdm.

        Args:
            opdm_aa: single spin sector of the 1-particle denstiy matrix

        Returns:  InteractionRDM object for full spin-orbital 1-RDM and 2-RDM
        """

        opdm = np.zeros((2 * self.num_orbitals, 2 * self.num_orbitals),
                        dtype=np.complex128)
        opdm[::2, ::2] = opdm_aa
        opdm[1::2, 1::2] = opdm_aa
        tpdm = linalg.wedge(opdm, opdm, (1, 1), (1, 1))
        rdms = InteractionRDM(opdm, 2 * tpdm)
        return rdms

    def energy_from_rhf_opdm(self, opdm_aa: np.ndarray) -> float:
        """
        Compute the energy given a spin-up opdm

        Args:
            opdm_aa: spin-up opdm.  Should be an n x n matrix where n is
                     the number of spatial orbitals

        Returns: RHF energy
        """
        rdms = self.rdms_from_rhf_opdm(opdm_aa)
        return rdms.expectation(self.hamiltonian).real

    def rhf_global_gradient(self, params: np.ndarray, alpha_opdm: np.ndarray):
        """
        Compute rhf global gradient

        Args:
            params: rhf-parameters for rotation matrix.
            alpha_opdm: 1-RDM corresponding to results of basis rotation
                        parameterized by `params'.

        Returns: gradient vector the same size as the input `params'
        """
        opdm = np.zeros((2 * self.num_orbitals, 2 * self.num_orbitals),
                        dtype=np.complex128)
        opdm[::2, ::2] = alpha_opdm
        opdm[1::2, 1::2] = alpha_opdm
        tpdm = 2 * linalg.wedge(opdm, opdm, (1, 1), (1, 1))

        # now go through and generate all the necessary Z, Y, Y_kl matrices
        kappa_matrix = rhf_params_to_matrix(params,
                                            len(self.occ) + len(self.virt),
                                            self.occ, self.virt)
        kappa_matrix_full = np.kron(kappa_matrix, np.eye(2))
        w_full, v_full = np.linalg.eigh(
            -1j * kappa_matrix_full)  # so that kappa = i U lambda U^
        eigs_scaled_full = get_matrix_of_eigs(w_full)

        grad = np.zeros(self.nocc * self.nvirt, dtype=np.complex128)
        # kdelta = np.eye(2 * self.num_orbitals)

        # NOW GENERATE ALL TERMS ASSOCIATED WITH THE GRADIENT!!!!!!
        for p in range(self.nocc * self.nvirt):
            grad_params = np.zeros_like(params)
            grad_params[p] = 1
            Y = rhf_params_to_matrix(grad_params,
                                     len(self.occ) + len(self.virt), self.occ,
                                     self.virt)
            Y_full = np.kron(Y, np.eye(2))

            # Now rotate Y int othe basis that diagonalizes Z
            Y_kl_full = v_full.conj().T.dot(Y_full).dot(v_full)
            # now rotate Y_{kl} * (exp(i(l_{k} - l_{l})) - 1) / (i(l_{k} - l_{l}))
            # into the original basis
            pre_matrix_full = v_full.dot(eigs_scaled_full * Y_kl_full).dot(
                v_full.conj().T)

            grad_expectation = -1.0 * np.einsum(
                'ab,pa,pb',
                self.hamiltonian.one_body_tensor,
                pre_matrix_full,
                opdm,
                optimize='optimal').real

            grad_expectation += 1.0 * np.einsum(
                'ab,bq,aq',
                self.hamiltonian.one_body_tensor,
                pre_matrix_full,
                opdm,
                optimize='optimal').real

            grad_expectation += 1.0 * np.einsum(
                'ijkl,pi,jpkl',
                self.hamiltonian.two_body_tensor,
                pre_matrix_full,
                tpdm,
                optimize='optimal').real

            grad_expectation += -1.0 * np.einsum(
                'ijkl,pj,ipkl',
                self.hamiltonian.two_body_tensor,
                pre_matrix_full,
                tpdm,
                optimize='optimal').real

            grad_expectation += -1.0 * np.einsum(
                'ijkl,kq,ijlq',
                self.hamiltonian.two_body_tensor,
                pre_matrix_full,
                tpdm,
                optimize='optimal').real

            grad_expectation += 1.0 * np.einsum(
                'ijkl,lq,ijkq',
                self.hamiltonian.two_body_tensor,
                pre_matrix_full,
                tpdm,
                optimize='optimal').real
            grad[p] = grad_expectation

        return grad


def generate_hamiltonian(one_body_integrals: np.ndarray,
                         two_body_integrals: np.ndarray,
                         constant: float,
                         EQ_TOLERANCE: Optional[float] = 1.0E-12
                        ) -> InteractionOperator:
    n_qubits = 2 * one_body_integrals.shape[0]
    # Initialize Hamiltonian coefficients.
    one_body_coefficients = np.zeros((n_qubits, n_qubits))
    two_body_coefficients = np.zeros((n_qubits, n_qubits, n_qubits, n_qubits))
    # Loop through integrals.
    for p in range(n_qubits // 2):
        for q in range(n_qubits // 2):

            # Populate 1-body coefficients. Require p and q have same spin.
            one_body_coefficients[2 * p, 2 * q] = one_body_integrals[p, q]
            one_body_coefficients[2 * p + 1, 2 * q +
                                  1] = one_body_integrals[p, q]
            # Continue looping to prepare 2-body coefficients.
            for r in range(n_qubits // 2):
                for s in range(n_qubits // 2):
                    # Mixed spin
                    two_body_coefficients[2 * p, 2 * q + 1, 2 * r + 1, 2 *
                                          s] = (two_body_integrals[p, q, r, s] /
                                                2.)
                    two_body_coefficients[2 * p + 1, 2 * q, 2 * r, 2 * s +
                                          1] = (two_body_integrals[p, q, r, s] /
                                                2.)

                    # Same spin
                    two_body_coefficients[2 * p, 2 * q, 2 * r, 2 *
                                          s] = (two_body_integrals[p, q, r, s] /
                                                2.)
                    two_body_coefficients[2 * p + 1, 2 * q + 1, 2 * r +
                                          1, 2 * s +
                                          1] = (two_body_integrals[p, q, r, s] /
                                                2.)

    # Truncate.
    one_body_coefficients[
        np.absolute(one_body_coefficients) < EQ_TOLERANCE] = 0.
    two_body_coefficients[
        np.absolute(two_body_coefficients) < EQ_TOLERANCE] = 0.

    # Cast to InteractionOperator class and return.
    molecular_hamiltonian = InteractionOperator(constant, one_body_coefficients,
                                                two_body_coefficients)
    return molecular_hamiltonian


def rhf_params_to_matrix(parameters: np.ndarray,
                         num_orbitals: int,
                         occ: Optional[Union[None, List[int]]] = None,
                         virt: Optional[Union[None, List[int]]] = None
                        ) -> np.ndarray:
    """
    For restricted Hartree-Fock we have nocc * nvirt parameters.  These are
    provided as a list that is ordered by (virtuals) \times (occupied).

    For example, for H4 we have 2 orbitals occupied and 2 virtuals

    occupied = [0, 1]  virtuals = [2, 3]

    parameters = [(v_{0}, o_{0}), (v_{0}, o_{1}), (v_{1}, o_{0}), (v_{1}, o_{1})]
               = [(2, 0), (2, 1), (3, 0), (3, 1)]

    You can think of the tuples of elements of the upper right triangle of the
    antihermitian matrix that specifies the c_{b, i} coefficients.

    coefficient matrix
    [[ c_{0, 0}, -c_{1, 0}, -c_{2, 0}, -c_{3, 0}],
     [ c_{1, 0},  c_{1, 1}, -c_{2, 1}, -c_{3, 1}],
     [ c_{2, 0},  c_{2, 1},  c_{2, 2}, -c_{3, 2}],
     [ c_{3, 0},  c_{3, 1},  c_{3, 2},  c_{3, 3}]]

    Since we are working with only non-redundant operators we know c_{i, i} = 0
    and any c_{i, j} where i and j are both in occupied or both in virtual = 0.

    Args:
        parameters: array of parameters for kappa matrix
        num_orbitals: total number of spatial orbitals
        occ: (Optional) indices for doubly occupied sector
        virt: (Optional) indices for virtual sector

    Returns: np.ndarray kappa matrix
    """
    if occ is None:
        occ = range(num_orbitals // 2)
    if virt is None:
        virt = range(num_orbitals // 2, num_orbitals)

    # check that parameters are a real array
    if not np.allclose(parameters.imag, 0):
        raise ValueError("parameters input must be real valued")

    kappa = np.zeros((len(occ) + len(virt), len(occ) + len(virt)))
    for idx, (v, o) in enumerate(product(virt, occ)):
        kappa[v, o] = parameters[idx].real
        kappa[o, v] = -parameters[idx].real
    return kappa


def rhf_func_generator(rhf_func: HartreeFockFunctional,
                       init_occ_vec: Optional[Union[None, np.ndarray]] = None,
                       get_opdm_func: Optional[bool] = False
                      ) -> Union[Tuple[Callable, Callable, Callable],
                                 Tuple[Callable, Callable, Callable, Callable]]:
    """
    Generate the energy, gradient, and unitary functions

    Args:
        rhf_func: objective function object.
        init_occ_vec: (optional) vector for occupation numbers of
                      the alpha-opdm.
        get_opdm_func: (optional) flag for returning Callable that returns
                       the final opdm.
    Returns: functions for unitary, energy, gradient (in that order)
    """
    if init_occ_vec is None:
        initial_opdm = np.diag([1] * rhf_func.nocc + [0] * rhf_func.nvirt)
    else:
        initial_opdm = np.diag(init_occ_vec)

    def energy(params):
        u = unitary(params)
        final_opdm_aa = u.dot(initial_opdm).dot(np.conjugate(u).T)
        tenergy = rhf_func.energy_from_rhf_opdm(final_opdm_aa)
        return tenergy

    def gradient(params):
        u = unitary(params)
        final_opdm_aa = u.dot(initial_opdm).dot(np.conjugate(u).T)
        return rhf_func.rhf_global_gradient(params, final_opdm_aa).real

    def unitary(params):
        kappa = rhf_params_to_matrix(params, rhf_func.nocc + rhf_func.nvirt,
                                     rhf_func.occ, rhf_func.virt)
        return sp.linalg.expm(kappa)

    def get_opdm(params):
        u = unitary(params)
        return u.dot(initial_opdm).dot(np.conjugate(u).T)

    if get_opdm_func:
        return unitary, energy, gradient, get_opdm
    return unitary, energy, gradient


def rhf_minimization(rhf_object: HartreeFockFunctional,
                     method: Optional[str] = 'CG',
                     initial_guess: Optional[Union[None, np.ndarray]] = None,
                     verbose: Optional[bool] = True,
                     sp_options: Optional[Union[None, Dict]] = None
                    ) -> OptimizeResult:
    """
    Perform Hartree-Fock energy minimization

    Args:
        rhf_object: An instantiation of the HartreeFockFunctional
        method: (optional) scipy optimization method
        initial_guess: (optional) initial rhf parameter vector.  If None
                       zero vector is used.
        verbose: (optional) turn on printing.  This is passed to the
                 scipy 'disp' option.
        sp_options:
    Returns: scipy.optimize result object
    """
    _, energy, gradient = rhf_func_generator(rhf_object)
    if initial_guess is None:
        init_guess = np.zeros(rhf_object.nocc * rhf_object.nvirt)
    else:
        init_guess = np.asarray(initial_guess).flatten()

    sp_optimizer_options = {'disp': verbose}
    if sp_options is not None:
        sp_optimizer_options.update(sp_options)

    return sp.optimize.minimize(energy,
                                init_guess,
                                jac=gradient,
                                method=method,
                                options=sp_optimizer_options)



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Tests for vpe_estimators.py"""

import pytest
import numpy
import pandas
import cirq

from .vpe_estimators import (
    PhaseFitEstimator,
    get_phase_function,
)

rng = numpy.random.RandomState(seed=42)


def test_requests_simulation_at_pi_for_pauli():
    estimator = PhaseFitEstimator(evals=[-1, +1])
    sim_points = estimator.get_simulation_points(safe=False)
    assert len(sim_points) == 2
    assert numpy.isclose(sim_points[0], 0)
    assert numpy.isclose(sim_points[1], numpy.pi / 2)


def test_estimates_expectation_value_pauli_nonoise():
    evals = numpy.array([-1, +1])
    true_amps = numpy.array([0.2, 0.8])
    true_expectation_value = numpy.dot(evals, true_amps)

    estimator = PhaseFitEstimator(evals)
    sim_points = estimator.get_simulation_points()
    phase_function = numpy.array([
        numpy.sum([
            amp * numpy.exp(1j * ev * time)
            for ev, amp in zip(evals, true_amps)
        ])
        for time in sim_points
    ])
    print(phase_function)
    test_expectation_value = estimator.get_expectation_value(phase_function)
    assert numpy.isclose(true_expectation_value, test_expectation_value)


def test_estimates_expectation_value_scattered_nonoise():
    evals = rng.uniform(-1, 1, 10)
    true_amps = rng.uniform(0, 1, 10)
    true_amps = true_amps / numpy.sum(true_amps)
    true_expectation_value = numpy.dot(evals, true_amps)

    estimator = PhaseFitEstimator(evals)
    sim_points = estimator.get_simulation_points()
    phase_function = numpy.array([
        numpy.sum([
            amp * numpy.exp(1j * ev * time)
            for ev, amp in zip(evals, true_amps)
        ])
        for time in sim_points
    ])
    test_expectation_value = estimator.get_expectation_value(phase_function)

    assert numpy.isclose(true_expectation_value, test_expectation_value)


def test_phase_function_gen_raises_error():
    results = [0, 0]
    qubits = [cirq.GridQubit(0, 0), cirq.GridQubit(0, 1)]
    target_qid = 0
    with pytest.raises(ValueError):
        get_phase_function(results, qubits, target_qid)


def test_phase_function_gen():

    class FakeResult:

        def __init__(self, data):
            self.data = {'msmt': pandas.Series(data)}

    datasets = [[] for j in range(8)]
    for xindex in [2, 3, 6, 7]:
        datasets[xindex] = [0] * 50 + [2] * 50
    for z0index in [0, 4]:
        datasets[z0index] = [0] * 100
    for z1index in [1, 5]:
        datasets[z1index] = [2] * 100

    results = [FakeResult(dataset) for dataset in datasets]
    qubits = [cirq.GridQubit(0, 0), cirq.GridQubit(0, 1)]
    target_qid = 0
    phase_function_est = get_phase_function(results, qubits, target_qid)
    assert phase_function_est == 1



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
'''Tests for circuit_validation.py'''

import cirq
from openfermion import QubitOperator

from openfermion.testing.circuit_validation import\
    validate_trotterized_evolution


def test_checking_passes():
    '''Simple test that circuit checking passes'''
    angle = 0.652
    z_rotation_op = angle * QubitOperator('Z0')
    qubits = [cirq.GridQubit(0, 0)]
    circuit = cirq.Circuit([cirq.rz(-2 * angle).on(qubits[0])])
    res = validate_trotterized_evolution(circuit, [z_rotation_op], qubits)
    assert res is True


def test_checking_passes_twoops():
    '''Simple test that circuit checking passes'''
    anglez = 0.652
    anglex = 0.334
    z_rotation_op = anglez * QubitOperator('Z0')
    x_rotation_op = anglex * QubitOperator('X0')
    qubits = [cirq.GridQubit(0, 0)]
    circuit = cirq.Circuit([
        cirq.rz(-2 * anglez).on(qubits[0]),
        cirq.rx(-2 * anglex).on(qubits[0])
    ])
    res = validate_trotterized_evolution(circuit,
                                         [z_rotation_op, x_rotation_op], qubits)
    assert res is True


def test_checking_fails():
    '''Simple test that circuit checking fails'''
    angle = 0.652
    z_rotation_op = angle * QubitOperator('Z0')
    qubits = [cirq.GridQubit(0, 0)]
    circuit = cirq.Circuit([cirq.rz(angle).on(qubits[0])])
    res = validate_trotterized_evolution(circuit, [z_rotation_op], qubits)
    assert res is False



from itertools import product
import numpy
from openfermion.ops.representations import InteractionOperator


def make_reduced_hamiltonian(molecular_hamiltonian: InteractionOperator,
                             n_electrons: int) -> InteractionOperator:
    r"""
    Construct the reduced Hamiltonian.

    This Hamiltonian is equivalent to the electronic structure Hamiltonian
    but contains only two-body terms.  To do this, the operator now depends
    on the number of particles being simulated.  We use the RDM sum rule to
    lift the 1-body terms to the two-body space.

    Derivation:
        use the fact that i^l = (1/(n -1)) sum_{jk}\delta_{jk}i^ j^ k l
                          i^l = (-1/(n -1)) sum_{jk}\delta_{jk}j^ i^ k l
                          i^l = (-1/(n -1)) sum_{jk}\delta_{jk}i^ j^ l k
                          i^l = (1/(n -1)) sum_{jk}\delta_{jk}j^ i^ l k

        Rewrite each one-body term as an even weighting of all four 2-RDM
        elements with delta functions. Then rearrange terms so that each ijkl
        term gets a sum of permuted one-body terms multiplied by delta
        function. One should notice that this results in the same formula
        if one was to apply the wedge product!

    Args:
        molecular_hamiltonian: operator to write reduced hamiltonian for
        n_electrons: number of electrons in the system
    Returns:
        InteractionOperator with a zero one-body component.
    """
    constant = molecular_hamiltonian.constant
    h1 = molecular_hamiltonian.one_body_tensor
    h2 = molecular_hamiltonian.two_body_tensor

    delta = numpy.eye(h1.shape[0])
    k2 = numpy.zeros_like(h2)
    normalization = 1 / (4 * (n_electrons - 1))
    for i, j, k, l in product(range(h1.shape[0]), repeat=4):
        k2[i, j, k, l] = normalization * (
            h1[i, l] * delta[j, k] + h1[j, k] * delta[i, l] -
            h1[i, k] * delta[j, l] - h1[j, l] * delta[i, k]) + h2[i, j, k, l]

    return InteractionOperator(constant, numpy.zeros_like(h1), k2)



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from .gates import (
    FSWAP,
    FSwapPowGate,
    Rxxyy,
    Ryxxy,
    Rzz,
    rot11,
    state_swap_eigen_component,
    fermionic_simulation_gates_from_interaction_operator,
    sum_of_interaction_operator_gate_generators,
    ParityPreservingFermionicGate,
    InteractionOperatorFermionicGate,
    QuadraticFermionicSimulationGate,
    CubicFermionicSimulationGate,
    QuarticFermionicSimulationGate,
    DoubleExcitation,
    DoubleExcitationGate,
    rot111,
    CRxxyy,
    CRyxxy,
)

from .lcu_util import (
    preprocess_lcu_coefficients_for_reversible_sampling,
    lambda_norm,
)

from .low_rank import (
    get_chemist_two_body_coefficients,
    low_rank_two_body_decomposition,
    prepare_one_body_squared_evolution,
)

from .primitives import (
    bogoliubov_transform,
    ffft,
    optimal_givens_decomposition,
    prepare_gaussian_state,
    prepare_slater_determinant,
    swap_network,
)

from .slater_determinants import (
    gaussian_state_preparation_circuit,
    slater_determinant_preparation_circuit,
    jw_get_gaussian_state,
    jw_slater_determinant,
)

from .trotter_exp_to_qgates import (
    trotter_operator_grouping,
    pauli_exp_to_qasm,
    trotterize_exp_qubop_to_qasm,
)

from .unitary_cc import (
    uccsd_generator,
    uccsd_convert_amplitude_format,
    uccsd_singlet_paramsize,
    uccsd_singlet_get_packed_amplitudes,
    uccsd_singlet_generator,
)

from .trotter import (
    LINEAR_SWAP_NETWORK,
    LinearSwapNetworkTrotterAlgorithm,
    LOW_RANK,
    LowRankTrotterAlgorithm,
    SplitOperatorTrotterAlgorithm,
    SplitOperatorTrotterStep,
    SymmetricSplitOperatorTrotterStep,
    ControlledAsymmetricSplitOperatorTrotterStep,
    AsymmetricSplitOperatorTrotterStep,
    ControlledSymmetricSplitOperatorTrotterStep,
    SPLIT_OPERATOR,
    diagonal_coulomb_potential_and_kinetic_terms_as_arrays,
    bit_mask_of_modes_acted_on_by_fermionic_terms,
    split_operator_trotter_error_operator_diagonal_two_body,
    fermionic_swap_trotter_error_operator_diagonal_two_body,
    simulation_ordered_grouped_hubbard_terms_with_info,
    low_depth_second_order_trotter_error_operator,
    low_depth_second_order_trotter_error_bound,
    simulation_ordered_grouped_low_depth_terms_with_info,
    stagger_with_info,
    simulate_trotter,
    TrotterAlgorithm,
    TrotterStep,
    error_bound,
    error_operator,
    trotter_steps_required,
)

from .vpe_circuits import (
    vpe_single_circuit,
    vpe_circuits_single_timestep,
    standard_vpe_rotation_set,
)



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""This module contains functions for compiling circuits to prepare
Slater determinants and fermionic Gaussian states."""

import numpy

from openfermion.config import EQ_TOLERANCE
from openfermion.ops import QuadraticHamiltonian
from openfermion.linalg.givens_rotations import (
    fermionic_gaussian_decomposition, givens_decomposition)
from openfermion.linalg.sparse_tools import (
    jw_configuration_state, jw_sparse_givens_rotation,
    jw_sparse_particle_hole_transformation_last_mode)


def gaussian_state_preparation_circuit(quadratic_hamiltonian,
                                       occupied_orbitals=None,
                                       spin_sector=None):
    r"""Obtain the description of a circuit which prepares a fermionic Gaussian
    state.

    Fermionic Gaussian states can be regarded as eigenstates of quadratic
    Hamiltonians. If the Hamiltonian conserves particle number, then these are
    just Slater determinants. See arXiv:1711.05395 for a detailed description
    of how this procedure works.

    The circuit description is returned as a sequence of elementary
    operations; operations that can be performed in parallel are grouped
    together. Each elementary operation is either

    - the string 'pht', indicating the particle-hole transformation
      on the last fermionic mode, which is the operator :math:`\mathcal{B}`
      such that

      .. math::

          \begin{align}
              \mathcal{B} a_N \mathcal{B}^\dagger &= a_N^\dagger,\\
              \mathcal{B} a_j \mathcal{B}^\dagger &= a_j, \quad
                  j = 1, \ldots, N-1,
          \end{align}

      or

    - a tuple :math:`(i, j, \theta, \varphi)`, indicating the operation

      .. math::
          \exp[i \varphi a_j^\dagger a_j]
          \exp[\theta (a_i^\dagger a_j - a_j^\dagger a_i)],

      a Givens rotation of modes :math:`i` and :math:`j` by angles
      :math:`\theta` and :math:`\varphi`.

    Args:
        quadratic_hamiltonian(QuadraticHamiltonian):
            The Hamiltonian whose eigenstate is desired.
        occupied_orbitals(list):
            A list of integers representing the indices of the occupied
            orbitals in the desired Gaussian state. If this is None
            (the default), then it is assumed that the ground state is
            desired, i.e., the orbitals with negative energies are filled.
        spin_sector (optional str): An optional integer specifying
            a spin sector to restrict to: 0 for spin-up and 1 for
            spin-down. If specified, the returned circuit acts on modes
            indexed by spatial indices (rather than spin indices).
            Should only be specified if the Hamiltonian
            includes a spin degree of freedom and spin-up modes
            do not interact with spin-down modes.

    Returns
    -------
        circuit_description (list[tuple]):
            A list of operations describing the circuit. Each operation
            is a tuple of objects describing elementary operations that
            can be performed in parallel. Each elementary operation
            is either the string 'pht', indicating a particle-hole
            transformation on the last fermionic mode, or a tuple of
            the form :math:`(i, j, \theta, \varphi)`,
            indicating a Givens rotation
            of modes :math:`i` and :math:`j` by angles :math:`\theta`
            and :math:`\varphi`.
        start_orbitals (list):
            The occupied orbitals to start with. This describes the
            initial state that the circuit should be applied to: it should
            be a Slater determinant (in the computational basis) with these
            orbitals filled.
    """
    if not isinstance(quadratic_hamiltonian, QuadraticHamiltonian):
        raise ValueError('Input must be an instance of QuadraticHamiltonian.')

    orbital_energies, transformation_matrix, _ = (
        quadratic_hamiltonian.diagonalizing_bogoliubov_transform(
            spin_sector=spin_sector))

    if quadratic_hamiltonian.conserves_particle_number:
        if occupied_orbitals is None:
            # The ground state is desired, so we fill the orbitals that have
            # negative energy
            occupied_orbitals = numpy.where(orbital_energies < 0.0)[0]

        # Get the unitary rows which represent the Slater determinant
        slater_determinant_matrix = transformation_matrix[occupied_orbitals]

        # Get the circuit description
        circuit_description = slater_determinant_preparation_circuit(
            slater_determinant_matrix)
        start_orbitals = range(len(occupied_orbitals))
    else:
        # TODO implement this
        if spin_sector is not None:
            raise NotImplementedError("Not yet supported")
        # Rearrange the transformation matrix because the circuit generation
        # routine expects it to describe annihilation operators rather than
        # creation operators
        n_qubits = quadratic_hamiltonian.n_qubits
        left_block = transformation_matrix[:, :n_qubits]
        right_block = transformation_matrix[:, n_qubits:]
        # Can't use numpy.block because that requires numpy>=1.13.0
        new_transformation_matrix = numpy.empty((n_qubits, 2 * n_qubits),
                                                dtype=complex)
        new_transformation_matrix[:, :n_qubits] = numpy.conjugate(right_block)
        new_transformation_matrix[:, n_qubits:] = numpy.conjugate(left_block)

        # Get the circuit description
        decomposition, left_decomposition, _, _ = (
            fermionic_gaussian_decomposition(new_transformation_matrix))
        if occupied_orbitals is None:
            # The ground state is desired, so the circuit should be applied
            # to the vaccuum state
            start_orbitals = []
            circuit_description = list(reversed(decomposition))
        else:
            start_orbitals = occupied_orbitals
            # The circuit won't be applied to the ground state, so we need to
            # use left_decomposition
            circuit_description = list(
                reversed(decomposition + left_decomposition))

    return circuit_description, start_orbitals


def slater_determinant_preparation_circuit(slater_determinant_matrix):
    r"""Obtain the description of a circuit which prepares a Slater determinant.

    The input is an :math:`N_f \times N` matrix :math:`Q` with orthonormal
    rows. Such a matrix describes the Slater determinant

    .. math::

        b^\dagger_1 \cdots b^\dagger_{N_f} \lvert \text{vac} \rangle,

    where

    .. math::

        b^\dagger_j = \sum_{k = 1}^N Q_{jk} a^\dagger_k.

    The output is the description of a circuit which prepares this
    Slater determinant, up to a global phase.
    The starting state which the circuit should be applied to
    is a Slater determinant (in the computational basis) with
    the first :math:`N_f` orbitals filled.

    Args:
        slater_determinant_matrix: The matrix :math:`Q` which describes the
            Slater determinant to be prepared.
    Returns:
        circuit_description:
            A list of operations describing the circuit. Each operation
            is a tuple of elementary operations that can be performed in
            parallel. Each elementary operation is a tuple of the form
            :math:`(i, j, \theta, \varphi)`, indicating a Givens rotation
            of modes :math:`i` and :math:`j` by angles :math:`\theta`
            and :math:`\varphi`.
    """
    decomposition, _, _ = givens_decomposition(slater_determinant_matrix)
    circuit_description = list(reversed(decomposition))
    return circuit_description


def jw_get_gaussian_state(quadratic_hamiltonian, occupied_orbitals=None):
    """Compute an eigenvalue and eigenstate of a quadratic Hamiltonian.

    Eigenstates of a quadratic Hamiltonian are also known as fermionic
    Gaussian states.

    Args:
        quadratic_hamiltonian(QuadraticHamiltonian):
            The Hamiltonian whose eigenstate is desired.
        occupied_orbitals(list):
            A list of integers representing the indices of the occupied
            orbitals in the desired Gaussian state. If this is None
            (the default), then it is assumed that the ground state is
            desired, i.e., the orbitals with negative energies are filled.

    Returns
    -------
        energy (float):
            The eigenvalue.
        state (sparse):
            The eigenstate in scipy.sparse csc format.
    """
    if not isinstance(quadratic_hamiltonian, QuadraticHamiltonian):
        raise ValueError('Input must be an instance of QuadraticHamiltonian.')

    n_qubits = quadratic_hamiltonian.n_qubits

    # Compute the energy
    orbital_energies, constant = quadratic_hamiltonian.orbital_energies()
    if occupied_orbitals is None:
        # The ground energy is desired
        if quadratic_hamiltonian.conserves_particle_number:
            num_negative_energies = numpy.count_nonzero(
                orbital_energies < -EQ_TOLERANCE)
            occupied_orbitals = range(num_negative_energies)
        else:
            occupied_orbitals = []
    energy = numpy.sum(orbital_energies[occupied_orbitals]) + constant

    # Obtain the circuit that prepares the Gaussian state
    circuit_description, start_orbitals = \
        gaussian_state_preparation_circuit(quadratic_hamiltonian,
                                           occupied_orbitals)

    # Initialize the starting state
    state = jw_configuration_state(start_orbitals, n_qubits)

    # Apply the circuit
    if not quadratic_hamiltonian.conserves_particle_number:
        particle_hole_transformation = (
            jw_sparse_particle_hole_transformation_last_mode(n_qubits))
    for parallel_ops in circuit_description:
        for op in parallel_ops:
            if op == 'pht':
                state = particle_hole_transformation.dot(state)
            else:
                i, j, theta, phi = op
                state = jw_sparse_givens_rotation(i, j, theta, phi,
                                                  n_qubits).dot(state)

    return energy, state


def jw_slater_determinant(slater_determinant_matrix):
    r"""Obtain a Slater determinant.

    The input is an :math:`N_f \times N` matrix :math:`Q` with orthonormal
    rows. Such a matrix describes the Slater determinant

    .. math::

        b^\dagger_1 \cdots b^\dagger_{N_f} \lvert \text{vac} \rangle,

    where

    .. math::

        b^\dagger_j = \sum_{k = 1}^N Q_{jk} a^\dagger_k.

    Args:
        slater_determinant_matrix: The matrix :math:`Q` which describes the
            Slater determinant to be prepared.
    Returns:
        The Slater determinant as a sparse matrix.
    """
    circuit_description = slater_determinant_preparation_circuit(
        slater_determinant_matrix)
    start_orbitals = range(slater_determinant_matrix.shape[0])
    n_qubits = slater_determinant_matrix.shape[1]

    # Initialize the starting state
    state = jw_configuration_state(start_orbitals, n_qubits)

    # Apply the circuit
    for parallel_ops in circuit_description:
        for op in parallel_ops:
            i, j, theta, phi = op
            state = jw_sparse_givens_rotation(i, j, theta, phi,
                                              n_qubits).dot(state)

    return state
